import argparse
from datetime import datetime
from processing import process_file_pair
import json
import numpy as np
import os
from os import path
import file_handling as fh

class MineException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ConfigException(Exception):
    def __init__(self, message):
        super().__init__(message)


default_options = {
    "use_time": False,
    "run_shuffle": False,
    "th_test": np.sqrt(0.5),
    "taylor_sig": 0.05,
    "taylor_cut": 0.1,
    "th_lax": 0.8,
    "th_sqr": 0.5,
    "history": 10.0,
    "taylor_look": 0.5,
    "jacobian": False,
    "n_epochs": 100,
    "miner_verbose": False,
    "miner_train_fraction": 2.0/3
}


if __name__ == '__main__':

    # the following will prevent tensorflow from using the GPU - as the used models have very low complexity
    # they will generally be fit faster on the CPU - furthermore parallelization currently used
    # will not work if tensorflow is run on the GPU!!
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

    a_parser = argparse.ArgumentParser(prog="process_csv.py",
                                       description="Uses MINE to fit and interpret CNN models that relate predictors"
                                                   "identified by one CSV file to responses identified by another.")
    a_parser.add_argument("-p", "--predictors", help="Path to CSV files of predictors or alternatively "
                                                     "directory with predictor files.",
                          type=str, required=True, nargs='+')
    a_parser.add_argument("-r", "--responses", help="Path to CSV files of responses or alternatively "
                                                    "directory with response files.",
                          type=str, required=True, nargs='+')
    a_parser.add_argument("-ut", "--use_time", help="If set time will be used as one predictor.",
                          action='store_true')
    a_parser.add_argument("-sh", "--run_shuffle", help="If set shuffled controls will be run as well.",
                          action='store_true')
    a_parser.add_argument("-ct", "--th_test", help="The test score threshold to "
                                                   "decide that fit was successful.",
                          type=float, default=default_options['th_test'])
    a_parser.add_argument("-ts", "--taylor_sig", help="The significance threshold for taylor expansion.",
                          type=float, default=default_options['taylor_sig'])
    a_parser.add_argument("-tc", "--taylor_cut", help="The variance fraction that has to be lost to"
                                                      "consider component important for fit.",
                          type=float, default=default_options['taylor_cut'])
    a_parser.add_argument("-la", "--th_lax", help="The threshold of variance explained by the linear"
                                                      "approximation to consider the fit linear.",
                          type=float, default=default_options['th_lax'])
    a_parser.add_argument("-lsq", "--th_sqr", help="The threshold of variance explained by the 2nd order"
                                                  "approximation to consider the fit 2nd order.",
                          type=float, default=default_options['th_sqr'])
    a_parser.add_argument("-n", "--model_name", help="Name of model for file saving purposes.", type=str)
    a_parser.add_argument("-mh", "--history", help="The length of model history in time units.",
                          type=float, default=default_options['history'])
    a_parser.add_argument("-tl", "--taylor_look", help="Determines taylor look ahead as multiplier of history",
                          type=float, default=default_options['taylor_look'])
    a_parser.add_argument("-j", "--jacobian", help="Store the Jacobians (linear receptive fields) for each response.",
                          action='store_true')
    a_parser.add_argument("-o", "--config", help="Path to config file with run parameters.", type=str, default=None)
    a_parser.add_argument("-e", "--n_epochs", help="Number of epochs when fitting model.", type=int,
                          default=default_options['n_epochs'])
    a_parser.add_argument("-mv","--miner_verbose", help="Receive updates on model fitting in command line",
                          action='store_true')
    a_parser.add_argument("-mtf", "--miner_train_fraction", help="The fraction of data to use for training",
                          type=float, default=default_options['miner_train_fraction'])

    args = a_parser.parse_args()

    r_paths = fh.process_file_args(args.responses)
    p_paths = fh.process_file_args(args.predictors)

    file_pairs = fh.pair_files(r_paths, p_paths)

    config_dict = None
    if args.config is not None:
        if not path.exists(args.config):
            raise ConfigException("Config file does not exist")
        if not path.isfile(args.config):
            raise ConfigException("Config path is not a file")
        try:
            config_dict = json.load(open(args.config))["config"]
        except json.decoder.JSONDecodeError or UnicodeDecodeError:
            raise ConfigException("Config file does not contain valid JSON")
        except KeyError:
            raise ConfigException("Config file does not contain config section")
    else:
        # set to default options
        config_dict = default_options

    # any argument given on the command line will supersede corresponding options in the config dict
    time_as_pred = config_dict["use_time"] if args.use_time == default_options["use_time"] else args.use_time
    run_shuffle = config_dict["run_shuffle"] if args.run_shuffle == default_options["run_shuffle"] else args.run_shuffle
    test_score_thresh = config_dict["th_test"] if np.isclose(args.th_test, default_options["th_test"]) else args.th_test
    taylor_sig = config_dict["taylor_sig"] if np.isclose(args.taylor_sig, default_options["taylor_sig"]) else args.taylor_sig
    taylor_cutoff = config_dict["taylor_cut"] if np.isclose(args.taylor_cut, default_options["taylor_cut"]) else args.taylor_cut
    lax_thresh = config_dict["th_lax"] if np.isclose(args.th_lax, default_options["th_lax"]) else args.th_lax
    sqr_thresh = config_dict["th_sqr"] if np.isclose(args.th_sqr, default_options["th_sqr"]) else args.th_sqr
    history_time = config_dict["history"] if np.isclose(args.history, default_options["history"]) else args.history
    taylor_look_fraction = config_dict["taylor_look"] if np.isclose(args.taylor_look, default_options["taylor_look"]) else args.taylor_look
    fit_jacobian = config_dict["jacobian"] if args.jacobian == default_options["jacobian"] else args.jacobian
    fit_epochs = config_dict["n_epochs"] if args.n_epochs == default_options["n_epochs"] else args.n_epochs
    miner_train_fraction = config_dict["miner_train_fraction"] if args.miner_train_fraction == default_options["miner_train_fraction"] else args.miner_train_fraction
    miner_verbose = config_dict["miner_verbose"] if args.miner_verbose == default_options["miner_verbose"] else args.miner_verbose

    if args.model_name is None:
        # set to default to file name of predictors
        your_model = datetime.now().strftime("%B_%d_%Y_%I_%M%p")
    else:
        your_model = args.model_name

    for i, pair in enumerate(file_pairs):
        # save run information and configuration used as json file which we set up here
        configuration = {
            "config":
                {
                    "use_time": time_as_pred,
                    "run_shuffle": run_shuffle,
                    "th_test": test_score_thresh,
                    "taylor_sig": taylor_sig,
                    "taylor_cut": taylor_cutoff,
                    "th_lax": lax_thresh,
                    "th_sqr": sqr_thresh,
                    "history": history_time,
                    "taylor_look": taylor_look_fraction,
                    "jacobian": fit_jacobian,
                    "n_epochs": fit_epochs,
                    "miner_verbose": miner_verbose,
                    "miner_train_fraction": miner_train_fraction
                },
            "run":
                {
                    "model_name": your_model if len(file_pairs)==1 else f"{your_model}_{i}",
                    "predictor_file": pair[1],
                    "response_file": pair[0],
                    "timestamp": datetime.now().now().isoformat(),
                }
        }


        ###
        # Load and process data
        ###
        process_file_pair(pair[0], pair[1], configuration)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 6.10.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QLabel, QLineEdit,
    QPushButton, QSizePolicy, QVBoxLayout, QWidget)
from neuro_mine.ui import resources_rc

class Ui_Widget(object):
    def setupUi(self, Widget):
        if not Widget.objectName():
            Widget.setObjectName(u"Widget")
        Widget.resize(604, 559)
        self.pushButton = QPushButton(Widget)
        self.pushButton.setObjectName(u"pushButton")
        self.pushButton.setGeometry(QRect(510, 530, 91, 32))
        self.pushButton_2 = QPushButton(Widget)
        self.pushButton_2.setObjectName(u"pushButton_2")
        self.pushButton_2.setGeometry(QRect(500, 40, 81, 32))
        self.pushButton_3 = QPushButton(Widget)
        self.pushButton_3.setObjectName(u"pushButton_3")
        self.pushButton_3.setGeometry(QRect(500, 70, 81, 32))
        self.pushButton_4 = QPushButton(Widget)
        self.pushButton_4.setObjectName(u"pushButton_4")
        self.pushButton_4.setGeometry(QRect(510, 500, 91, 32))
        self.pushButton_5 = QPushButton(Widget)
        self.pushButton_5.setObjectName(u"pushButton_5")
        self.pushButton_5.setGeometry(QRect(410, 120, 171, 32))
        self.widget = QWidget(Widget)
        self.widget.setObjectName(u"widget")
        self.widget.setGeometry(QRect(260, 0, 251, 551))
        self.verticalLayout = QVBoxLayout(self.widget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.lineEdit = QLineEdit(self.widget)
        self.lineEdit.setObjectName(u"lineEdit")

        self.verticalLayout.addWidget(self.lineEdit)

        self.lineEdit_4 = QLineEdit(self.widget)
        self.lineEdit_4.setObjectName(u"lineEdit_4")

        self.verticalLayout.addWidget(self.lineEdit_4)

        self.lineEdit_2 = QLineEdit(self.widget)
        self.lineEdit_2.setObjectName(u"lineEdit_2")

        self.verticalLayout.addWidget(self.lineEdit_2)

        self.checkBox = QCheckBox(self.widget)
        self.checkBox.setObjectName(u"checkBox")

        self.verticalLayout.addWidget(self.checkBox)

        self.checkBox_2 = QCheckBox(self.widget)
        self.checkBox_2.setObjectName(u"checkBox_2")

        self.verticalLayout.addWidget(self.checkBox_2)

        self.lineEdit_3 = QLineEdit(self.widget)
        self.lineEdit_3.setObjectName(u"lineEdit_3")

        self.verticalLayout.addWidget(self.lineEdit_3)

        self.lineEdit_5 = QLineEdit(self.widget)
        self.lineEdit_5.setObjectName(u"lineEdit_5")

        self.verticalLayout.addWidget(self.lineEdit_5)

        self.lineEdit_11 = QLineEdit(self.widget)
        self.lineEdit_11.setObjectName(u"lineEdit_11")

        self.verticalLayout.addWidget(self.lineEdit_11)

        self.lineEdit_6 = QLineEdit(self.widget)
        self.lineEdit_6.setObjectName(u"lineEdit_6")

        self.verticalLayout.addWidget(self.lineEdit_6)

        self.lineEdit_7 = QLineEdit(self.widget)
        self.lineEdit_7.setObjectName(u"lineEdit_7")

        self.verticalLayout.addWidget(self.lineEdit_7)

        self.lineEdit_8 = QLineEdit(self.widget)
        self.lineEdit_8.setObjectName(u"lineEdit_8")

        self.verticalLayout.addWidget(self.lineEdit_8)

        self.checkBox_4 = QCheckBox(self.widget)
        self.checkBox_4.setObjectName(u"checkBox_4")

        self.verticalLayout.addWidget(self.checkBox_4)

        self.lineEdit_9 = QLineEdit(self.widget)
        self.lineEdit_9.setObjectName(u"lineEdit_9")

        self.verticalLayout.addWidget(self.lineEdit_9)

        self.lineEdit_12 = QLineEdit(self.widget)
        self.lineEdit_12.setObjectName(u"lineEdit_12")

        self.verticalLayout.addWidget(self.lineEdit_12)

        self.lineEdit_13 = QLineEdit(self.widget)
        self.lineEdit_13.setObjectName(u"lineEdit_13")

        self.verticalLayout.addWidget(self.lineEdit_13)

        self.checkBox_3 = QCheckBox(self.widget)
        self.checkBox_3.setObjectName(u"checkBox_3")

        self.verticalLayout.addWidget(self.checkBox_3)

        self.lineEdit_10 = QLineEdit(self.widget)
        self.lineEdit_10.setObjectName(u"lineEdit_10")

        self.verticalLayout.addWidget(self.lineEdit_10)

        self.pushButton_6 = QPushButton(Widget)
        self.pushButton_6.setObjectName(u"pushButton_6")
        self.pushButton_6.setGeometry(QRect(360, 530, 141, 32))
        self.layoutWidget = QWidget(Widget)
        self.layoutWidget.setObjectName(u"layoutWidget")
        self.layoutWidget.setGeometry(QRect(0, 10, 261, 521))
        self.verticalLayout_2 = QVBoxLayout(self.layoutWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.label_2 = QLabel(self.layoutWidget)
        self.label_2.setObjectName(u"label_2")

        self.verticalLayout_2.addWidget(self.label_2)

        self.label_3 = QLabel(self.layoutWidget)
        self.label_3.setObjectName(u"label_3")

        self.verticalLayout_2.addWidget(self.label_3)

        self.label_4 = QLabel(self.layoutWidget)
        self.label_4.setObjectName(u"label_4")

        self.verticalLayout_2.addWidget(self.label_4)

        self.label_5 = QLabel(self.layoutWidget)
        self.label_5.setObjectName(u"label_5")

        self.verticalLayout_2.addWidget(self.label_5)

        self.label_6 = QLabel(self.layoutWidget)
        self.label_6.setObjectName(u"label_6")

        self.verticalLayout_2.addWidget(self.label_6)

        self.label_7 = QLabel(self.layoutWidget)
        self.label_7.setObjectName(u"label_7")

        self.verticalLayout_2.addWidget(self.label_7)

        self.label_8 = QLabel(self.layoutWidget)
        self.label_8.setObjectName(u"label_8")

        self.verticalLayout_2.addWidget(self.label_8)

        self.label_13 = QLabel(self.layoutWidget)
        self.label_13.setObjectName(u"label_13")

        self.verticalLayout_2.addWidget(self.label_13)

        self.label_9 = QLabel(self.layoutWidget)
        self.label_9.setObjectName(u"label_9")

        self.verticalLayout_2.addWidget(self.label_9)

        self.label_10 = QLabel(self.layoutWidget)
        self.label_10.setObjectName(u"label_10")

        self.verticalLayout_2.addWidget(self.label_10)

        self.label_11 = QLabel(self.layoutWidget)
        self.label_11.setObjectName(u"label_11")

        self.verticalLayout_2.addWidget(self.label_11)

        self.label_14 = QLabel(self.layoutWidget)
        self.label_14.setObjectName(u"label_14")

        self.verticalLayout_2.addWidget(self.label_14)

        self.label_12 = QLabel(self.layoutWidget)
        self.label_12.setObjectName(u"label_12")

        self.verticalLayout_2.addWidget(self.label_12)

        self.label_16 = QLabel(self.layoutWidget)
        self.label_16.setObjectName(u"label_16")

        self.verticalLayout_2.addWidget(self.label_16)

        self.label_17 = QLabel(self.layoutWidget)
        self.label_17.setObjectName(u"label_17")

        self.verticalLayout_2.addWidget(self.label_17)

        self.label_18 = QLabel(self.layoutWidget)
        self.label_18.setObjectName(u"label_18")

        self.verticalLayout_2.addWidget(self.label_18)

        self.label_15 = QLabel(self.layoutWidget)
        self.label_15.setObjectName(u"label_15")

        self.verticalLayout_2.addWidget(self.label_15)

        self.label_19 = QLabel(Widget)
        self.label_19.setObjectName(u"label_19")
        self.label_19.setGeometry(QRect(500, 160, 58, 16))
        self.label_20 = QLabel(Widget)
        self.label_20.setObjectName(u"label_20")
        self.label_20.setGeometry(QRect(500, 200, 58, 16))
        self.label_21 = QLabel(Widget)
        self.label_21.setObjectName(u"label_21")
        self.label_21.setGeometry(QRect(500, 230, 58, 16))
        self.label_22 = QLabel(Widget)
        self.label_22.setObjectName(u"label_22")
        self.label_22.setGeometry(QRect(500, 260, 58, 16))
        self.label_23 = QLabel(Widget)
        self.label_23.setObjectName(u"label_23")
        self.label_23.setGeometry(QRect(500, 290, 58, 16))
        self.label_24 = QLabel(Widget)
        self.label_24.setObjectName(u"label_24")
        self.label_24.setGeometry(QRect(500, 330, 58, 16))
        self.label_25 = QLabel(Widget)
        self.label_25.setObjectName(u"label_25")
        self.label_25.setGeometry(QRect(500, 390, 58, 16))
        self.label_26 = QLabel(Widget)
        self.label_26.setObjectName(u"label_26")
        self.label_26.setGeometry(QRect(500, 420, 58, 16))
        self.label_27 = QLabel(Widget)
        self.label_27.setObjectName(u"label_27")
        self.label_27.setGeometry(QRect(500, 450, 58, 16))
        self.label = QLabel(Widget)
        self.label.setObjectName(u"label")
        self.label.setGeometry(QRect(530, 10, 41, 21))
        self.label.setPixmap(QPixmap(u":/logo.png"))
        self.label_28 = QLabel(Widget)
        self.label_28.setObjectName(u"label_28")
        self.label_28.setGeometry(QRect(580, 50, 16, 16))
        self.label_28.setPixmap(QPixmap(u":/question.png"))
        self.label_29 = QLabel(Widget)
        self.label_29.setObjectName(u"label_29")
        self.label_29.setGeometry(QRect(580, 80, 16, 16))
        self.label_29.setPixmap(QPixmap(u":/question.png"))
        self.pushButton.raise_()
        self.pushButton_2.raise_()
        self.pushButton_3.raise_()
        self.pushButton_4.raise_()
        self.widget.raise_()
        self.pushButton_6.raise_()
        self.layoutWidget.raise_()
        self.label_19.raise_()
        self.label_20.raise_()
        self.label_21.raise_()
        self.label_22.raise_()
        self.label_23.raise_()
        self.label_24.raise_()
        self.label_25.raise_()
        self.label_26.raise_()
        self.label_27.raise_()
        self.label.raise_()
        self.label_28.raise_()
        self.label_29.raise_()
        self.pushButton_5.raise_()

        self.retranslateUi(Widget)

        QMetaObject.connectSlotsByName(Widget)
    # setupUi

    def retranslateUi(self, Widget):
        Widget.setWindowTitle(QCoreApplication.translate("Widget", u"Neuro MINE", None))
        self.pushButton.setText(QCoreApplication.translate("Widget", u"Run Model", None))
        self.pushButton_2.setText(QCoreApplication.translate("Widget", u"Browse...", None))
        self.pushButton_3.setText(QCoreApplication.translate("Widget", u"Browse...", None))
        self.pushButton_4.setText(QCoreApplication.translate("Widget", u"Browse...", None))
        self.pushButton_5.setText(QCoreApplication.translate("Widget", u"Restore Preset Parameters", None))
        self.checkBox.setText(QCoreApplication.translate("Widget", u"Yes", None))
        self.checkBox_2.setText(QCoreApplication.translate("Widget", u"Yes", None))
        self.checkBox_4.setText(QCoreApplication.translate("Widget", u"Yes", None))
        self.checkBox_3.setText(QCoreApplication.translate("Widget", u"Yes", None))
        self.pushButton_6.setText(QCoreApplication.translate("Widget", u"Save Configuration", None))
        self.label_2.setText(QCoreApplication.translate("Widget", u"Model Name:", None))
        self.label_3.setText(QCoreApplication.translate("Widget", u"Predictor File Path:", None))
#if QT_CONFIG(tooltip)
        self.label_4.setToolTip(QCoreApplication.translate("Widget", u"Single delimiter-separated file with Time as first column", None))
#endif // QT_CONFIG(tooltip)
        self.label_4.setText(QCoreApplication.translate("Widget", u"Response File Path:", None))
        self.label_5.setText(QCoreApplication.translate("Widget", u"Use Time as a Predictor:", None))
        self.label_6.setText(QCoreApplication.translate("Widget", u"Run Shuffle:", None))
        self.label_7.setText(QCoreApplication.translate("Widget", u"Test Score Threshold:", None))
        self.label_8.setText(QCoreApplication.translate("Widget", u"Taylor Expansion Significance Threshold:", None))
        self.label_13.setText(QCoreApplication.translate("Widget", u"Taylor Expansion Look Ahead:", None))
        self.label_9.setText(QCoreApplication.translate("Widget", u"Taylor Cutoff", None))
        self.label_10.setText(QCoreApplication.translate("Widget", u"Linear Fit Variance Fraction:", None))
        self.label_11.setText(QCoreApplication.translate("Widget", u"Square Fit Variance Fraction:", None))
        self.label_14.setText(QCoreApplication.translate("Widget", u"Store Linear Receptive Fields (Jacobians):", None))
        self.label_12.setText(QCoreApplication.translate("Widget", u"Model History (seconds):", None))
        self.label_16.setText(QCoreApplication.translate("Widget", u"Number of Epochs:", None))
        self.label_17.setText(QCoreApplication.translate("Widget", u"Train Data Fraction:", None))
        self.label_18.setText(QCoreApplication.translate("Widget", u"Verbose:", None))
        self.label_15.setText(QCoreApplication.translate("Widget", u"Upload JSON File with Existing Parameters:", None))
        self.label_19.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label_20.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label_21.setText(QCoreApplication.translate("Widget", u"(0,4)", None))
        self.label_22.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label_23.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label_24.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label_25.setText(QCoreApplication.translate("Widget", u"\u22651", None))
        self.label_26.setText(QCoreApplication.translate("Widget", u"[0,100]", None))
        self.label_27.setText(QCoreApplication.translate("Widget", u"[0,1]", None))
        self.label.setText("")
#if QT_CONFIG(tooltip)
        self.label_28.setToolTip(QCoreApplication.translate("Widget", u"Single delimiter-separated file with Time as first column", None))
#endif // QT_CONFIG(tooltip)
        self.label_28.setText("")
#if QT_CONFIG(tooltip)
        self.label_29.setToolTip(QCoreApplication.translate("Widget", u"Single delimiter-separated file with Time as first column", None))
#endif // QT_CONFIG(tooltip)
        self.label_29.setText("")
    # retranslateUi





"""
Module with data preparation classes and functions
"""
from __future__ import annotations

import numpy as np
import tensorflow as tf
import h5py
from typing import Union, List, Any, Optional
from numba import njit
from warnings import warn


def create_overwrite(storage: Union[h5py.File, h5py.Group], name: str, data: Any, overwrite: bool,
                     compress=False) -> None:
    """
    Allows to create a new dataset in an hdf5 file an if desired overvwrite any old data
    :param storage: The hdf5 file or group used to store the information
    :param name: The name of the dataset
    :param data: The data
    :param overwrite: If true any old data with the same name will be deleted and subsequently replaced
    :param compress: If true, data will be stored compressed
    """
    if overwrite and name in storage:
        del storage[name]
    if compress:
        storage.create_dataset(name, data=data, compression="gzip", compression_opts=5)
    else:
        storage.create_dataset(name, data=data)


def modelweights_to_hdf5(storage: Union[h5py.File, h5py.Group], m_weights: List[np.ndarray]) -> None:
    """
    Stores tensorflow weights of a model sequentially to an hdf5 file or group to allow for more compact storage
    across many models (NOTE: No states are saved)
    :param storage: The hdf5 file or group used to store the information
    :param m_weights: List of weight arrays returned by keras.model.get_weights()
    """
    storage.create_dataset("n_layers", data=len(m_weights))
    for i, mw in enumerate(m_weights):
        if type(mw) == np.ndarray:
            storage.create_dataset(f"layer_{i}", data=mw, compression="gzip", compression_opts=5)
        else:
            storage.create_dataset(f"layer_{i}", data=mw)


def modelweights_from_hdf5(storage: Union[h5py.File, h5py.Group]) -> List[np.ndarray]:
    """
    Loads tensorflow model weights from an hdf5 file or group
    :param storage: The hdf5 file or group from which to load the information
    :return: List of weight arrays that can be used with keras.model.set_weights() to set model weights
    """
    n_layers = storage["n_layers"][()]
    m_weights = []
    for i in range(n_layers):
        m_weights.append(storage[f"layer_{i}"][()])
    return m_weights


def bootstrap_fractional_r2loss(real: np.ndarray, predicted: np.ndarray, remainder: np.ndarray,
                                n_boot: int) -> np.ndarray:
    """
    Returns bootstrap samples for the loss in r^2 after components are removed from a prediction
    :param real: The real timeseries data
    :param predicted: The full prediction for the timeseries
    :param remainder: The prediction for the timeseries after components have been excluded
    :param n_boot:
    :return: n_boot long vector of fractional loss scores (fraction of r2 that is lost going from full to remainder)
    """
    if real.size != predicted.size or predicted.size != remainder.size:
        raise ValueError("All timeseries inputs must have same length")
    if n_boot <= 1:
        raise ValueError("n_boot must be > 1")
    output = np.empty(n_boot)
    indices = np.arange(real.size)
    # bootstrap loop
    for i in range(n_boot):
        choose = np.random.choice(indices, indices.size, replace=True)
        full_r2 = np.corrcoef(real[choose], predicted[choose])[0, 1]**2
        rem_r2 = np.corrcoef(real[choose], remainder[choose])[0, 1]**2
        output[i] = 1 - rem_r2 / full_r2
    return output


def bootstrap_binned_average(x: np.ndarray, weights: np.ndarray, bins: np.ndarray, n_boot: int) -> np.ndarray:
    """
    Returns bootstrap samples for a weighted histogram (i.e. average of a quantity based on bins of another quantity)
    :param x: The quantity to bin by
    :param weights: The quantity to average
    :param bins: Array of bin boundaries
    :param n_boot: The number of bootstrap samples to generate
    :return: n_boot x (bins.size-1) matrix of binned average bootstrap samples
    """
    if x.ndim > 1 or weights.ndim > 1:
        raise ValueError("Only 1D vectors are supported for x and weights")
    if x.size != weights.size:
        raise ValueError("Each x sample must have an associated weight")
    output = np.empty((n_boot, bins.size - 1))
    indices = np.arange(x.size)
    for i in range(n_boot):
        choose = np.random.choice(indices, indices.size, replace=True)
        x_bs = x[choose]
        w_bs = weights[choose]
        weighted_count = np.histogram(x_bs, bins, weights=w_bs)[0].astype(float)
        count = np.histogram(x_bs, bins)[0].astype(float)
        output[i, :] = weighted_count / count
    return output


def bootstrap(data: np.ndarray, nboot: int, bootfun: callable) -> np.ndarray:
    """
    For a a n_samples x m_features array creates nboot bootstrap variates of bootfun
    :param data: The data to be bootstrapped
    :param nboot: The number of boostrap variates to create
    :param bootfun: The function to apply, must take axis parameter
    :return: nboot x m_features array of bootstrap variates
    """
    indices = np.arange(data.shape[0]).astype(int)
    variates = np.full((nboot, data.shape[1]), np.nan)
    for i in range(nboot):
        chosen = np.random.choice(indices, data.shape[0], True)
        variates[i, :] = bootfun(data[chosen], axis=0)
    return variates


def safe_standardize(x: np.ndarray, axis: Optional[int] = None, epsilon=1e-9) -> np.ndarray:
    """
    Standardizes an array to 0 mean and unit standard deviation avoiding division by 0
    :param x: The array to standardize
    :param axis: The axis along which standardization should be performmed
    :param epsilon: Small constant to add to standard deviation to avoid divide by 0 if sd(x)=0
    :return: The standardized array of same dimension as x
    """
    if x.ndim == 1 or axis is None:
        y = x - np.mean(x)
        y /= (np.std(y) + epsilon)
    else:
        y = x - np.mean(x, axis=axis, keepdims=True)
        y /= (np.std(y, axis=axis, keepdims=True) + epsilon)
    return y


def barcode_cluster(x: np.ndarray, threshold: Union[float, np.ndarray]) -> np.ndarray:
    """
    For n_samples by n_features input matrix assigns a cluster to each member based on "barcoding"
    where all above-threshold features are set to contribute to a sample
    :param x: n_samples x n_features input
    :param threshold: The trehshold(s) above which (all) features contribute - either scalar or n_features vector
    :return: n_samples long vector of cluster numbers. Ordered as if contributions were binary digits with index 0
        of x having highest significance (no contribution would be first, all contributing last)
    """

    def split(m, row_ix=None, index=0):
        if row_ix is None:
            row_ix = np.arange(m.shape[0])
        above = m[m[:, index] > 0]
        rix_above = row_ix[m[:, index] > 0]
        below = m[m[:, index] <= 0]
        rix_below = row_ix[m[:, index] <= 0]
        if above.size == 0 and index == m.shape[1]-1:
            return [rix_below]
        if below.size == 0 and index == m.shape[1]-1:
            return [rix_above]
        if index == m.shape[1]-1:
            return [rix_above, rix_below]
        if above.size == 0:
            return split(below, rix_below, index+1)
        if below.size == 0:
            return split(above, rix_above, index+1)
        return split(above, rix_above, index+1) + split(below, rix_below, index+1)

    if not np.isscalar(threshold):
        if threshold.size != x.shape[1]:
            raise ValueError("Threshold either has to be a scalar or a vector with n_features element")
        threshold = threshold.ravel()[None, :]
    xt = x > threshold
    clustered_indices = split(xt)
    cluster_numbers = np.full(x.shape[0], np.nan)
    for i, clust in enumerate(reversed(clustered_indices)):
        cluster_numbers[clust] = i
    return cluster_numbers


def create_coordinate_grid_points(x: np.ndarray, y: np.ndarray, z: Optional[np.ndarray]) -> np.ndarray:
    """
    Creates linearized grid of all coordinate pairs
    :param x: Vector of x coordinates in the grid
    :param y: Vector of y coordinates in the grid
    :param z: None or vector of z coordinates in the grid
    :return: n x 2 or n x 3 matrix of all points in the grid
    """
    if z is None:
        xv, yv = np.meshgrid(x, y)
        xv = xv.reshape((np.prod(xv.shape),))
        yv = yv.reshape((np.prod(yv.shape),))
        return np.hstack((xv[:, None], yv[:, None]))
    else:
        xv, yv, zv = np.meshgrid(x, y, z)
        xv = xv.reshape((np.prod(xv.shape),))
        yv = yv.reshape((np.prod(yv.shape),))
        zv = zv.reshape((np.prod(zv.shape),))
        return np.hstack((xv[:, None], yv[:, None], zv[:, None]))


@njit
def rearrange_hessian(hessian: np.ndarray, npreds: int, inp_length: int) -> np.ndarray:
    """
    Re-arranges contents of our hessian matrices so that consecutive rows/columns are grouped by predictor
    instead of by time
    :param hessian: The hessian
    :param npreds: The number of predictors
    :param inp_length: The timelength of each regressor input
    :return: The re-arranged hessian
    """
    hessian_r = np.empty_like(hessian, dtype=np.float32)
    for row in range(hessian.shape[0]):
        regnum = row % npreds
        time = row // npreds
        row_ix = regnum * inp_length + time
        for col in range(hessian.shape[1]):
            regnum = col % npreds
            time = col // npreds
            col_ix = regnum * inp_length + time
            hessian_r[row_ix, col_ix] = hessian[row, col]
    return hessian_r


def simulate_response(act_predictor, predictors: np.ndarray) -> np.ndarray:
    """
    Simulate the predicted response of a neuron to an arbitrary input
    :param act_predictor: The model used to predict the response
    :param predictors: n_time x m_predictors matrix of predictor inputs
    :return: n_time - history_length + 1 long vector of predicted neural responses
    """
    history = act_predictor.input_length
    pred = [act_predictor.get_output(predictors[None, t - history + 1:t + 1, :]) for t in
            range(history - 1, predictors.shape[0])]
    return np.hstack(pred)


def modified_gram_schmidt(col_mat: np.ndarray) -> np.ndarray:
    """
    Performs orthogonalization of col_mat such that in case of linear dependence, linearly
    dependent columns will be set to all 0
    :param col_mat: mxn matrix with columns containing features
    :return: mxn matrix with orthogonalized columns
    """
    # initialize with copy
    v = col_mat.copy()
    # iterate through all columns
    for j in range(v.shape[1]):
        # if the current column is linearly dependent to previous columns
        # its values will be close to 0 - we set to 0 exactly and move on
        if np.allclose(v[:, j], 0):
            v[:, j] = np.zeros(v.shape[0])
            continue
        n = np.linalg.norm(v[:, j])
        q = v[:, j] / n  # this is the unit vector we will  project out all *subsequent* columns one-by-one
        for k in range(j+1, v.shape[1]):
            v[:, k] = v[:, k] - (q@v[:, k])*q
    # set vector lengths to unit norm (in a safe manner avoiding div-0 for 0 vectors)
    norms = np.linalg.norm(v, axis=0, keepdims=True)
    norms[norms < 1e-9] = 1e-9
    v /= norms
    return v


def sigmoid(x):
    return 1/(1 + np.exp(-x))


def interp_events(x: np.ndarray, xp: np.ndarray, fp: np.ndarray) -> np.ndarray:
    """
    Interpolate spiking trace adding spikes to the correct intervals without generating in-between values
    :param x: The desired timepoints after interpolation
    :param xp: The timepoints original timepoints
    :param fp: The spike data before interpolation
    """
    if xp.size != fp.size:
        raise ValueError(f"xp and fp must have the same size but sizes are {xp.size} and {fp.size}")
    if x.min() < xp.min() or x.max() > xp.max():
        warn("Trying to extrapolate spike data", UserWarning)
    if not np.all(np.logical_or(fp==0, fp==1)):
        raise ValueError("Values in fp suggest that this is not 0/1 coded spike data")
    f = np.zeros_like(x)
    for x_, f_ in zip(xp, fp):
        # for each spike that is within the bounds of the interpolation time
        # find the closest x to assign it to
        if x_ < x.min():
            continue
        if x_ > x.max():
            break
        if f_ == 1:
            ix = np.argmin(np.abs(x - x_))
            if ix.size > 1:
                ix = ix[0]
            f[ix] += 1
    if np.any(f > 1):
        print(f"Interpolation times too coarse for data. {np.sum(f > 1)} timepoints in interpolated data correspond to"
              f" more than one spike")
    f[f > 1] = 1
    return f


class Data:
    def __init__(self, input_steps, regressors: list, ca_responses: np.ndarray, tsteps_for_train=-1):
        """
        Creates a new Data class
        :param input_steps: The number of regressor timesteps into the past to use to model the response
        :param regressors: List of regressors. Vectors are shared for all ca_responses, while matrices must have
            same shape as ca_responses
        :param ca_responses: n_responses x m_timesteps matrix of cell calcium responses
        :param tsteps_for_train: If negative use all samples for training if positive use first m samples only
        """
        self.data_len = ca_responses.shape[1]
        self.n_responses = ca_responses.shape[0]
        self.regressors = []
        for i, reg in enumerate(regressors):
            if reg.ndim > 2:
                raise ValueError(f"Regressor {i} has more than 2 dimensions")
            elif reg.ndim == 2:
                if reg.shape[0] != 1 and reg.shape[0] != self.n_responses:
                    raise ValueError(f"Regressor {i} is matrix but does not have same amount of samples "
                                     f"as ca_responses")
                if reg.shape[1] != self.data_len:
                    raise ValueError(f"Regressor {i} needs to have same amount of timesteps as ca_responses")
            else:
                if reg.size != self.data_len:
                    raise ValueError(f"Regressor {i} needs to have same amount of timesteps as ca_responses")
                reg = reg[None, :]  # augment shape
            self.regressors.append(reg)
        self.ca_responses = ca_responses
        self.input_steps = input_steps
        if tsteps_for_train > 0:
            self.tsteps_for_train = tsteps_for_train
        elif tsteps_for_train == 0:
            raise ValueError("tsteps_for_train has to be either negative or larger 0")
        else:
            self.tsteps_for_train = ca_responses.shape[1]

    def training_data(self, sample_ix: int, batch_size=32):
        """
        Creates training data for the indicated calcium response sample (cell)
        :param sample_ix: The index of the cell
        :param batch_size: The training batch size to use
        :return: Tensorflow dataset that can be used for training with randomization
        """
        out_data = self.ca_responses[sample_ix, self.input_steps-1:self.tsteps_for_train].copy()
        in_data = np.full((out_data.size, self.input_steps, len(self.regressors)), np.nan).astype(np.float32)
        for i, reg in enumerate(self.regressors):
            if reg.shape[0] == 1:
                this_reg = reg
            else:
                this_reg = reg[sample_ix, :][None, :]
            for t in range(self.input_steps-1, out_data.size+self.input_steps-1):
                in_data[t-self.input_steps+1, :, i] = this_reg[0, t-self.input_steps+1:t+1]
        train_ds = tf.data.Dataset.from_tensor_slices((in_data, out_data)).\
            shuffle(in_data.shape[0], reshuffle_each_iteration=True).batch(batch_size, drop_remainder=True)
        return train_ds.prefetch(tf.data.AUTOTUNE)

    def test_data(self, sample_ix: int, batch_size=32):
        """
        Creates test data for the indicated calcium response sample (cell)
        :param sample_ix: The index of the cell
        :param batch_size: The training batch size to use
        :return: Tensorflow dataset that can be used for testing
        """
        if self.tsteps_for_train == self.ca_responses.shape[1]:
            raise ValueError("All data is training data")
        out_data = self.ca_responses[sample_ix, self.tsteps_for_train+self.input_steps-1:].copy()
        in_data = np.full((out_data.size, self.input_steps, len(self.regressors)), np.nan).astype(np.float32)
        for i, reg in enumerate(self.regressors):
            if reg.shape[0] == 1:
                this_reg = reg
            else:
                this_reg = reg[sample_ix, :][None, :]
            for t in range(self.input_steps-1, out_data.size+self.input_steps-1):
                t_t = t + self.tsteps_for_train
                in_data[t-self.input_steps+1, :, i] = this_reg[0, t_t-self.input_steps+1:t_t+1]
        test_ds = tf.data.Dataset.from_tensor_slices((in_data, out_data)).batch(batch_size, drop_remainder=False)
        return test_ds.prefetch(tf.data.AUTOTUNE)

    def regressor_matrix(self, sample_ix: int) -> np.ndarray:
        """
        For a given sample returns regressor matrix
        :param sample_ix: The index of the cell
        :return: n_timesteps x m_regressors matrix of regressors for the given cell
        """
        reg_data = np.full((self.ca_responses.shape[1], len(self.regressors)), np.nan).astype(np.float32)
        for i, reg in enumerate(self.regressors):
            if reg.shape[0] == 1:
                this_reg = reg
            else:
                this_reg = reg[sample_ix, :]
            reg_data[:, i] = this_reg.ravel().copy()
        return reg_data

    def predict_response(self, sample_ix: int, act_predictor):
        """
        Obtains the predicted response for a given cell
        :param sample_ix: The index of the cell
        :param act_predictor: The model to perform the prediction
        :return:
            [0]: n_timesteps-input_steps+1 sized vector of response prediction
            [1]: Corresponding timesteps in the original calcium response
        """
        if act_predictor.input_length != self.input_steps:
            raise ValueError("Input length of activity prediction model and data class mismatch")
        regressors = self.regressor_matrix(sample_ix)
        pred = simulate_response(act_predictor, regressors)
        return pred, self.ca_responses[sample_ix, self.input_steps-1:]

    def subset(self, sample_indices: List[int]) -> Data:
        """
        Returns a subset of the data, copying  responses and predictors for the indicated cell indices
        :param sample_indices: List of cell indices which should be contained in the new data object
        :return: A data object with only the indicated subset of cells and associated predictors
        """
        new_regs = []
        for r in self.regressors:
            if r.shape[0] == 1:
                new_regs.append(r.copy())
            else:
                new_regs.append(r[sample_indices, :].copy())
        new_ca_responses = self.ca_responses[sample_indices, :].copy()
        return Data(self.input_steps, new_regs, new_ca_responses, self.tsteps_for_train)

    def save(self, filename: str, overwrite=False) -> None:
        """
        Serializes the data instance to an hdf5 file
        :param filename: The path and name of the hdf5 file to save to
        :param overwrite: If set to true and file already exists it will be overwritten if False and exists will fail
        """
        with h5py.File(filename, mode='w' if overwrite else 'x') as dfile:
            self.save_direct(dfile, overwrite)

    def save_direct(self, file: Union[h5py.File, h5py.Group], overwrite=False) -> None:
        """
        Serializes the data instance to an hdf5 file object or group in a file
        :param file: The hdf5 file object
        :param overwrite: If set to true and contents already exists will be overwritten if False and exists will fail
        """
        create_overwrite(file, "input_steps", self.input_steps, overwrite)
        create_overwrite(file, "tsteps_for_train", self.tsteps_for_train, overwrite)
        create_overwrite(file, "ca_responses", self.ca_responses, overwrite)
        create_overwrite(file, "n_regressors", len(self.regressors), overwrite)  # saved  for purposes of easy loading
        if "regressors" not in file:
            r_group = file.create_group("regressors")
        else:
            r_group = file["regressors"]
        for i, r in enumerate(self.regressors):
            create_overwrite(r_group, str(i), r, overwrite)

    @staticmethod
    def load(filename: str) -> Data:
        """
        Loads a stored data instance from an hdf5 file
        :param filename: The path and name of the hdf5 file containing the stored data instance
        :return: A Data object with the contents loaded from file
        """
        with h5py.File(filename, mode='r') as dfile:
            return Data.load_direct(dfile)

    @staticmethod
    def load_direct(file: Union[h5py.File, h5py.Group]) -> Data:
        """
        Loads a stored data instance directly from an hdf5 file object or group
        :param file: The hdf5 file object
        :return: A Data object with the contents loaded from file
        """
        input_steps = file["input_steps"][()]
        tsteps_for_train = file["tsteps_for_train"][()]
        ca_responses = file["ca_responses"][()]
        n_regressors = file["n_regressors"][()]
        r_group = file["regressors"]
        regressors = []
        for i in range(n_regressors):
            regressors.append(r_group[str(i)][()])
        return Data(input_steps, regressors, ca_responses, tsteps_for_train)


if __name__ == "__main__":
    print("Module with data preparation classes and functions")



"""
Module for decomposing the prediction ANN into piecewise linear functions
via Taylor Series decomposition
"""

import neuro_mine.scripts.model as model
import numpy as np
import tensorflow as tf
from numba import njit
from typing import List, Tuple
from neuro_mine.scripts import utilities


@tf.function
def d2ca_dr2(mdl: model.ActivityPredictor, reg_input: np.ndarray) -> Tuple[tf.Tensor, tf.Tensor]:
    x = tf.convert_to_tensor(reg_input)
    with tf.GradientTape() as t2:
        t2.watch(x)
        with tf.GradientTape() as t1:
            t1.watch(x)
            # NOTE: The following is slightly faster than ca = mdl(x) presumably due to skipping of dropout layers
            # However it is not compatible with resizing the model!!
            # c = mdl.conv_layer(x)
            # d1 = mdl.deep_1(mdl.flatten(c))
            # d2 = mdl.deep_2(d1)
            # ca = mdl.out(d2)
            ca = mdl(x)
        jacobian = t1.gradient(ca, x)
    hessian = t2.jacobian(jacobian, x)
    return jacobian, hessian


@tf.function
def dca_dr(mdl: model.ActivityPredictor, reg_input: np.ndarray) -> tf.Tensor:
    x = tf.convert_to_tensor(reg_input)
    with tf.GradientTape() as t1:
        t1.watch(x)
        # NOTE: The following is slightly faster than ca = mdl(x) presumably due to skipping of dropout layers
        # but not compatible with change layer number in the model
        # c = mdl.conv_layer(x)
        # d1 = mdl.deep_1(mdl.flatten(c))
        # d2 = mdl.deep_2(d1)
        # ca = mdl.out(d2)
        ca = mdl(x)
    jacobian = t1.gradient(ca, x)
    return jacobian


def taylor_predict(mdl: model.ActivityPredictor, regressors: np.ndarray, use_d2: bool, take_every: int,
                   predict_ahead=1) -> Tuple[np.ndarray, np.ndarray]:
    """
    For each time t in regressors, evaluates the model, computes the selected derivatives and
    then attempts to predict the model response at time t+1
    :param mdl: The model to use for predictions
    :param regressors: The 2D regressor matrix, n_timesteps x m_regressors
    :param use_d2: If set to False only the first derivative will be used for the prediction
    :param take_every: Only form predictions every n frames to save time
    :param predict_ahead: The number of frames to predict ahead with the taylor expansion
    :returns:
        [0]: (n_timesteps-input_length-predict_ahead)/n long timeseries of taylor predictions
        [1]: (n_timesteps-input_length-predict_ahead)/n long timeseries of actual network outputs
    """
    if predict_ahead < 1:
        raise ValueError("predict_ahead has to be integer >= 1")
    inp_length = mdl.input_length
    mdl_output = []
    taylor_prediction = []
    t = inp_length - 1
    while t < regressors.shape[0]-predict_ahead:
        cur_regs = regressors[None, t - inp_length + 1: t + 1, :]
        next_regs = regressors[None, t - inp_length + 1 + predict_ahead: t + 1 + predict_ahead, :]
        cur_mod_out = mdl.get_output(cur_regs)
        next_mod_out = mdl.get_output(next_regs)
        mdl_output.append(next_mod_out)
        if use_d2:
            d1, d2 = d2ca_dr2(mdl, cur_regs)
            d1 = d1.numpy().ravel()
            d2 = np.reshape(d2.numpy(), (regressors.shape[1] * mdl.input_length,
                                         regressors.shape[1] * mdl.input_length))
        else:
            d1 = dca_dr(mdl, cur_regs)
            d1 = d1.numpy().ravel()
            d2 = None
        tay_pred = _taylor_predict(cur_regs, next_regs, cur_mod_out, d1, d2)
        taylor_prediction.append(tay_pred)
        t += take_every
    return np.hstack(taylor_prediction), np.hstack(mdl_output)


def _taylor_predict(reg_fix_point: np.ndarray, reg_test: np.ndarray, ann_fix: float, d1: np.ndarray,
                    d2: np.ndarray) -> float:
    """
    Computes the taylor prediction about a point for another test point nearby
    :param reg_fix_point: The regressor input at the fix point where derivatives have been calculated
    :param reg_test: The regressor input for which to predict the ann response
    :param ann_fix: The output of the ann at reg_fix_point
    :param d1: The set of first order partial derivatives at reg_fix_point
    :param d2: The matrix of second order partial derivatives at reg_fix_point
    """
    diff = (reg_test - reg_fix_point).ravel()
    if d2 is None:
        return ann_fix + np.dot(diff, d1)
    return ann_fix + np.dot(diff, d1) + 0.5 * np.sum(np.dot(diff[:, None], diff[None, :]) * d2)


@njit
def _compute_taylor_d2(reg_diff: np.ndarray, d2: np.ndarray, nregs: int, inp_length: int) -> np.ndarray:
    """
    Computes responses belonging to the second derivative, rearranging by regressor
    instead of by time
    :param reg_diff: The difference in regressors as 2D (1 x (n_regs*n_timepoints)) vector
    :param d2: The hessian
    :param nregs: The number of regressors
    :param inp_length: The timelength of each regressor input
    :return: The second derivative contribution ((n_regs*n_timepoints) x (n_regs*n_timepoints))
    """
    taylor_d2_temp = 0.5 * np.dot(reg_diff, reg_diff.T) * d2  # this matrix is organized by time not by regressor
    taylor_d2 = np.empty_like(taylor_d2_temp, dtype=np.float32)
    for row in range(taylor_d2_temp.shape[0]):
        regnum = row % nregs
        time = row // nregs
        row_ix = regnum * inp_length + time
        for col in range(taylor_d2_temp.shape[1]):
            regnum = col % nregs
            time = col // nregs
            col_ix = regnum * inp_length + time
            taylor_d2[row_ix, col_ix] = taylor_d2_temp[row, col]
    return taylor_d2


@njit
def _compute_by_reg(taylor_d1: np.ndarray, taylor_d2: np.ndarray, nregs: int, inp_length: int) -> np.ndarray:
    """
    Aggregates derivative contributions by regressor
    :param taylor_d1: The first partial derivative contributions (by regressor and time vector)
    :param taylor_d2: The second partial derivative contributions (by regressor and time square matrix)
    :param nregs: The number of regressors
    :param inp_length: The number of timepoints
    :return: Contribution aggregated by regressor as array (1 xnregs x nregs) to account for possible interactions
    """
    by_reg = np.full((1, nregs, nregs), 0.0, dtype=np.float32)
    for r1 in range(nregs):
        for r2 in range(nregs):
            if r1 == r2:
                # these are the non-interacting parts which need to take d1 into account
                by_reg[0, r1, r2] += np.sum(taylor_d1[r1 * inp_length:(r1 + 1) * inp_length])
            by_reg[0, r1, r2] += np.sum(
                taylor_d2[r1 * inp_length:(r1 + 1) * inp_length, r2 * inp_length:(r2 + 1) * inp_length])
    return by_reg


def taylor_decompose(mdl: model.ActivityPredictor, regressors: np.ndarray, take_every: int, predict_ahead: int,
                     use_d2=True) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Uses taylor decomposition to predict changes in network output around chosen point using the
    all information as well as only the information corresponding to each regressor and their
    interactions terms
    :param mdl: The model to use for predictions
    :param regressors: The 2D regressor matrix, n_timesteps x m_regressors
    :param take_every: Only form predictions every n frames to save time
    :param predict_ahead: The number of frames to predict ahead with the taylor expansion
    :param use_d2: If set to false only the first derivative will be used in the taylor expansion
    :returns:
        [0]: The true change for each timepoint by going predict_ahead frames into the future
                (n_timesteps-input_length-predict_ahead)/n long vector
        [1]: The predicted change for the whole taylor series
                (n_timesteps-input_length-predict_ahead)/n long vector
        [2]: Array of predicted changes by regressors and their interactions
                (n_timesteps-input_length-predict_ahead)/n x n_regressors x n_regressors
    """
    if predict_ahead < 1:
        raise ValueError("predict_ahead has to be integer >= 1")
    inp_length = mdl.input_length
    mdl_out_change = []
    full_tp_change = []
    by_reg_tp_change = []
    t = inp_length - 1
    nregs = regressors.shape[1]
    while t < regressors.shape[0]-predict_ahead:
        cur_regs = regressors[None, t - inp_length + 1: t + 1, :]
        next_regs = regressors[None, t - inp_length + 1 + predict_ahead: t + 1 + predict_ahead, :]
        reg_diff = (next_regs - cur_regs).ravel().astype(np.float64)
        cur_mod_out = mdl.get_output(cur_regs)
        next_mod_out = mdl.get_output(next_regs)
        mdl_out_change.append(next_mod_out - cur_mod_out)
        # unfortunately, none of the following are contiguous by regressor but rather by timepoint
        # we therefore need to reshape taylor_d1 and taylor_d2 to allow for easy indexing below
        if use_d2:
            d1, d2 = d2ca_dr2(mdl, cur_regs)
            d1 = d1.numpy().ravel()
            d2 = np.reshape(d2.numpy(), (regressors.shape[1] * mdl.input_length,
                                         regressors.shape[1] * mdl.input_length))
            taylor_d1 = reg_diff * d1.astype(np.float64)
            taylor_d1 = np.reshape(taylor_d1, (inp_length, nregs)).T.ravel()
            taylor_d2 = _compute_taylor_d2(reg_diff[:, None], d2.astype(np.float64), nregs, inp_length)
        else:
            d1 = dca_dr(mdl, cur_regs)
            d1 = d1.numpy().ravel()
            taylor_d1 = reg_diff * d1.astype(np.float64)
            taylor_d1 = np.reshape(taylor_d1, (inp_length, nregs)).T.ravel()
            taylor_d2 = np.zeros((taylor_d1.size, taylor_d1.size))
        full_tp_change.append(np.sum(taylor_d1) + np.sum(taylor_d2))
        by_reg = _compute_by_reg(taylor_d1, taylor_d2, nregs, inp_length)
        by_reg_tp_change.append(by_reg)
        t += take_every
    return np.hstack(mdl_out_change), np.hstack(full_tp_change), np.vstack(by_reg_tp_change)


def data_mean_prediction(mdl: model.ActivityPredictor, x_bar, j_x_bar, h_x_bar, regressors: np.ndarray, take_every: int,
                         use_probability: bool):
    """
    Computes the prediction of responses based on a fixed Taylor expansion of the network around a specific point
    in our case taken to be the data mean
    :param mdl: The CNN model
    :param x_bar: The data mean (or any arbitrary fix point)
    :param j_x_bar: The jacobian of the model at x_bar
    :param h_x_bar: The hession of the model at x_bar
    :param regressors: The 2D regressor matrix, n_timesteps x m_regressors
    :param take_every: Only compute metrics every n frames to save time
    :param use_probability: If set to true, all outputs will be transformed to probabilities via sigmoid transform
    :return:
        [0]: The prediction of the CNN model
        [1]: The prediction of the 2nd order fixed-point expansion
        [2]: The prediction of a linear fixed-point expansion
    """
    # NOTE: We should speed this up - create 2nd-order design matrix from input and then directly treat this as a
    # regression problem, replacing the ugly while-loop with a matrix multiplication
    f_x_bar = mdl(x_bar)
    mean_prediction: List[float] = []
    mean_prediction_lin: List[float] = []
    mdl_out: List[float] = []
    inp_length = mdl.input_length
    t: int = inp_length - 1
    # prepare our first and second derivatives at the data mean
    d1 = j_x_bar.numpy().ravel()
    d2 = np.reshape(h_x_bar.numpy(), (regressors.shape[1] * mdl.input_length, regressors.shape[1] * mdl.input_length))
    while t < regressors.shape[0]:
        cur_regs = regressors[None, t - inp_length + 1: t + 1, :]
        # compute our difference to the data mean
        reg_diff = (cur_regs - x_bar).ravel().astype(np.float64)
        # get the actual model prediction at the current point and add to our return
        if use_probability:
            cur_mod_out = mdl.get_probability(cur_regs)
        else:
            cur_mod_out = mdl.get_output(cur_regs)
        mdl_out.append(cur_mod_out)
        # compute taylor decomposition around data mean
        mp_lin = f_x_bar + np.dot(reg_diff, d1)
        mp = mp_lin + 0.5 * np.sum(np.dot(reg_diff[:, None], reg_diff[None, :]) * d2)
        if use_probability:
            mean_prediction.append(utilities.sigmoid(mp))
            mean_prediction_lin.append(utilities.sigmoid(mp_lin))
        else:
            mean_prediction.append(mp)
            mean_prediction_lin.append(mp_lin)
        t += take_every
    return np.hstack(mdl_out), np.hstack(mean_prediction), np.hstack(mean_prediction_lin)


def complexity_scores(mdl: model.ActivityPredictor, x_bar, j_x_bar, h_x_bar, regressors: np.ndarray, take_every: int,
                      use_probability: bool):
    """
    Computes complexity scores - the squared correlation of a linear and a squared model around the data mean
    :param mdl: The CNN model
    :param x_bar: The data mean (or any arbitrary fix point)
    :param j_x_bar: The jacobian of the model at x_bar
    :param h_x_bar: The hession of the model at x_bar
    :param regressors: The 2D regressor matrix, n_timesteps x m_regressors
    :param take_every: Only compute metrics every n frames to save time
    :param use_probability: If set to true, all outputs will be transformed to probabilities via sigmoid transform
    :return:
        [0]: The R2 (coefficient of determination) of the linear 1st order approximation
        [1]: The R2 of the 2nd order approximation
    """
    true_model, order_2, order_1 = data_mean_prediction(mdl, x_bar, j_x_bar, h_x_bar, regressors, take_every,
                                                        use_probability)
    ss_tot = np.sum((true_model - np.mean(true_model))**2)
    lin_score = 1 - np.sum((true_model - order_1)**2)/ss_tot
    sq_score = 1 - np.sum((true_model - order_2)**2)/ss_tot
    return lin_score, sq_score


if __name__ == "__main__":
    print("Module for ANN Taylor decomposition")



import csv
from dateutil import parser as dateparser
import numpy as np
from os import path
from datetime import datetime, date, time
from typing import Optional, Tuple, List, Union
import os


def process_file_args(files_or_dir: List[str]) -> List[str]:
    """
    Processes file arguments, expanding directories if present
    :param files_or_dir: Path to file or directory or list of paths to files
    :return: List of file paths
    """
    # Note: We allow lists of files or a single file or a single directory but not lists containing directories
    if len(files_or_dir) == 1:
        if path.isdir(files_or_dir[0]):
            elements = os.listdir(files_or_dir[0])
            return [path.join(files_or_dir[0], e) for e in elements if path.isfile(path.join(files_or_dir[0], e)) and not e[0]=='.']
        elif path.isfile(files_or_dir[0]):
            return [files_or_dir[0]]
        else:
            raise ValueError("Unrecognized path: {}".format(files_or_dir))
    for f in files_or_dir:
        if not path.isfile(f):
            raise ValueError(f"{f} is not a path to a file. Note that if multiple arguments "
                             "are provided they have to be files")
    return files_or_dir


def pair_files(resp_files: List[str], pred_files: List[str]) -> List[Tuple[str, str]]:
    """
    Takes a list of response files and predictor files, matches them and then returns a list of matched tuples
    :param resp_files: List of response files
    :param pred_files: List of predictor files
    :return: Matched pairs
    """
    # At this moment we perform an extremely simple matching: We assume that the alphabetical order of predictor
    # and response files is the same
    if len(resp_files) != len(pred_files):
        raise ValueError(f"Cannot match {len(pred_files)} predictor files to {len(resp_files)} response files")
    resp_files = sorted(resp_files)
    pred_files = sorted(pred_files)
    return [(r, p) for r,p in zip(resp_files, pred_files)]


class FileParser:
    """
    Base class for file parser
    """
    def __init__(self, file_path):
        """
        Creates a new file parser object
        :param file_path: The path to the file to parse
        """
        if not path.exists(file_path):
            raise FileNotFoundError(f"File {file_path} not found")
        if not path.isfile(file_path):
            raise FileNotFoundError(f"File {file_path} is not a file")
        self.filename = file_path

    @staticmethod
    def _parse_datetime(s: str) -> Optional[datetime]:
        """
        Tries to parse a datetime from a string
        :param s: The string to interpret as time
        :return: A datetime object or None if not a datetime input
        """
        # Check if the item can be converted to a floating-point number - if yes we treat it as "not a date"
        # and therefore return None
        try:
            float(s)
            return None  # not a date but a float
        except ValueError:
            pass
        try:
            return dateparser.parse(s)
        except (ValueError, OverflowError):
            return None  # neither a float nor a date - should we throw an exception here instead of returning None?

class CSVParser(FileParser):
    """
    Parser for CSV files
    """
    def __init__(self, file_path: str, prefix: str="col"):
        """
        Creates a new CSV file parser object
        :param file_path: The path to the file to parse
        :param prefix: If no column names are provided in the file this prefix will be used to label columns
        """
        super().__init__(file_path)
        self.delimiter = CSVParser._find_delimiter(self.filename)
        self.prefix = prefix

    @staticmethod
    def _find_delimiter(filename):
        sniffer = csv.Sniffer()
        with open(filename) as fp:
            delimiter = sniffer.sniff(fp.read(-1)).delimiter
        return delimiter

    def load_data(self) -> Tuple[np.ndarray, bool, List]:
        """
        Loads the data from the file
        :return:
        """
        # Load as text-file, processing line-by-line
        with open(self.filename, "r") as f:
            lines = f.readlines()

        # Identify number of header rows - these can be rows that contain actual column headers
        # or rows with descriptive text. In other words we try to find the first true data-row here
        # using simple heuristics which will likely break if the user supplies pathological files
        skip = 0  # counter of how many rows to skip as header
        ncols = None  # the number of actual data columns
        header_row = None  # the column labels

        # Detect where numeric data starts
        for line in lines:
            parts = line.strip().split(self.delimiter)
            # the first column has to contain time, however this can be either floating points
            # counting seconds/minutes/etc or actual time objects
            first_col = self._parse_datetime(parts[0])

            if first_col is not None:  # datetime in first column
                try:
                    [float(x) for x in parts[1:] if x != ""]
                    ncols = len(parts)
                    break
                except ValueError:
                    skip += 1  # the row contained a non-numerical object and is therefore considered a header
            else:  # try full row as floats, since time should be numerical
                try:
                    [float(x) for x in parts if x != ""]
                    ncols = len(parts)
                    break
                except ValueError:
                    skip += 1

        # Try to find column header definition in skipped rows - this should be a row
        # that has the same number of columns as the data
        assert ncols is not None
        if skip > 0:
            for i in range(skip - 1, -1, -1):
                parts = lines[i].strip().split(self.delimiter)
                if len(parts) == ncols:
                    header_row = parts
                    break

        # assemble all lines that have a full column count
        numeric_part = [line.strip().split(self.delimiter) for line in lines[skip:]]
        numeric_part = [row for row in numeric_part if len(row) == ncols]

        data = []
        # determine how time is saved within the file
        first_val = self._parse_datetime(numeric_part[0][0])
        first_is_datetime = first_val is not None

        if first_is_datetime:
            # In the following we want to recode time into a count of seconds relative to a start time
            # since we do not know when predictors recording started relative to response acquisition
            # this start time has to be independent of the timestamp itself. However, it shouldn't be
            # so far in the past that we lose resolution. Arbitrarily, we chose the millennium
            # The better way would clearly be to find the minimum of predictor and response start times
            # and use that value but that can only happen after file matching
            t0 = datetime.combine(date(2000, 1, 1), time())
            for row in numeric_part:
                t = self._parse_datetime(row[0])
                ts = (t - t0).total_seconds()
                rest = [float(x) if x != "" else np.nan for x in row[1:]]
                data.append([ts] + rest)
        else:
            for row in numeric_part:
                data.append([float(x) if x != "" else np.nan for x in row])

        data = np.array(data, dtype=float)
        data_has_header = header_row is not None

        # Make sure that the first column can be interpreted as time
        if any(np.diff(data[:, 0]) <= 0):
            raise ValueError("First column in datafile should be time, but values aren't strictly increasing")

        if data_has_header:
            data_header = header_row
        else:
            ncols = data.shape[1]
            # The name of the first column is Time
            data_header = ["Time"]+[f"{self.prefix}_{i}" for i in range(ncols-1)]

        return data, data_has_header, data_header


"""
Module for easy running of MINE on user data
"""
import h5py
import numpy as np
from typing import List, Optional, Union
import neuro_mine.scripts.utilities as utilities
import neuro_mine.scripts.model as model
from neuro_mine.scripts.taylorDecomp import taylor_decompose, d2ca_dr2, complexity_scores
from dataclasses import dataclass
import warnings
from sklearn.metrics import roc_auc_score

@dataclass(frozen=True)
class _BaseData:
    """
    Class for shared MINE return values
    """
    taylor_scores: Optional[np.ndarray]
    taylor_true_change: Optional[np.ndarray]
    taylor_full_prediction: Optional[np.ndarray]
    taylor_by_predictor: Optional[np.ndarray]
    model_lin_approx_scores: Optional[np.ndarray]
    model_2nd_approx_scores: Optional[np.ndarray]
    jacobians: Optional[np.ndarray]
    hessians: Optional[np.ndarray]

    def save_to_hdf5(self, file_object: Union[h5py.File, h5py.Group], overwrite=False) -> None:
        """
        Saves all contents to a hdf5 file or group object
        :param file_object: The file/group to save to
        :param overwrite: If true will overwrite data in the file
        """
        if self.taylor_scores is not None:
            utilities.create_overwrite(file_object, "taylor_scores", self.taylor_scores, overwrite)
            utilities.create_overwrite(file_object, "taylor_true_change", self.taylor_true_change, overwrite)
            utilities.create_overwrite(file_object, "taylor_full_prediction", self.taylor_full_prediction,
                                       overwrite)
            utilities.create_overwrite(file_object, "taylor_by_predictor", self.taylor_by_predictor, overwrite)
        if self.model_lin_approx_scores is not None:
            utilities.create_overwrite(file_object, "model_lin_approx_scores", self.model_lin_approx_scores,
                                       overwrite)
        if self.model_2nd_approx_scores is not None:
            utilities.create_overwrite(file_object, "model_2nd_approx_scores", self.model_2nd_approx_scores,
                                       overwrite)
        if self.jacobians is not None:
            utilities.create_overwrite(file_object, "jacobians", self.jacobians, overwrite)
        if self.hessians is not None:
            utilities.create_overwrite(file_object, "hessians", self.hessians, overwrite)

@dataclass(frozen=True)
class MineData(_BaseData):
    """Class for the return values of MINE"""
    correlations_trained: np.ndarray
    correlations_test: np.ndarray

    def save_to_hdf5(self, file_object: Union[h5py.File, h5py.Group], overwrite=False) -> None:
        """
        Saves all contents to a hdf5 file or group object
        :param file_object: The file/group to save to
        :param overwrite: If true will overwrite data in the file
        """
        super().save_to_hdf5(file_object, overwrite)
        utilities.create_overwrite(file_object, "correlations_trained", self.correlations_trained, overwrite)
        utilities.create_overwrite(file_object, "correlations_test", self.correlations_test, overwrite)


@dataclass(frozen=True)
class MineSpikingData(_BaseData):
    """Class for the return values of MINE after spiking analysis"""
    roc_auc_trained: np.ndarray
    roc_auc_test: np.ndarray

    def save_to_hdf5(self, file_object: Union[h5py.File, h5py.Group], overwrite=False) -> None:
        """
        Saves all contents to a hdf5 file or group object
        :param file_object: The file/group to save to
        :param overwrite: If true will overwrite data in the file
        """
        super().save_to_hdf5(file_object, overwrite)
        utilities.create_overwrite(file_object, "roc_auc_trained", self.roc_auc_trained, overwrite)
        utilities.create_overwrite(file_object, "roc_auc_test", self.roc_auc_test, overwrite)


class MineWarning(Warning):
    def __init__(self, message: str):
        super().__init__(message)


class Mine:
    """
    Class that collects intended model data and provides
    analysis function to be run on user-data
    """
    def __init__(self, train_fraction: float, model_history: int, score_cut: float, compute_taylor: bool,
                 return_jacobians: bool, taylor_look_ahead: int, taylor_pred_every: int, fit_spikes: bool):
        """
        Create a new Mine object
        :param train_fraction: The fraction of frames to use for training 0 < train <= 1
        :param model_history: The number of frames to include in the model "history" (Note 1)
        :param score_cut: Minimum correlation required on test data to compute other metrics
        :param compute_taylor: If true, perform taylor expansion and complexity/nonlin evaluation and return results
        :param return_jacobians: If true, return the model jacobians at the data mean
        :param taylor_look_ahead: How many frames into the future to compute the taylor expansion (usually a few secs)
        :param taylor_pred_every: Every how many frames to compute the taylor expansion to save time
        :param fit_spikes: If true, fit the spiking model
        """
        # Note 1: The ANN model purely looks into the past. However, the convolutional filters
        # can be centered arbitrarily by shifting predictor and response frames relative to
        # each other. See <processMusall.py> for an example
        if train_fraction <= 0 or train_fraction > 1:
            raise ValueError(f"train_fraction must be larger 0 and smaller or equal to 1 not {train_fraction}")
        self.train_fraction = train_fraction
        if model_history < 0:
            raise ValueError("model_history cant be < 0")
        self.model_history = model_history
        self.compute_taylor = compute_taylor
        self.return_jacobians = return_jacobians
        # set following to true to return hessians -
        # Note memory requirements of (n_sample x (n_timepointsxn_predictors)^2) sized array
        self.return_hessians = False
        # set the following to a hdf5 file our group object to store model-weights in subgroups labeled
        # according to "cell_{cell_index}_weights". These model-weights can be loaded into a list compatible
        # with tensorflow.keras.model.set_weights() using the utilities.modelweights_from_hdf5 function
        # NOTE: Before setting the weights, the model has to be initialized to the appropriate model structure
        # by evaluation on an appropriately structured test input
        self.model_weight_store: Union[None, h5py.Group, h5py.File] = None
        self.n_epochs = 100  # sensible default
        self.taylor_look_ahead = taylor_look_ahead
        self.taylor_pred_every = taylor_pred_every
        self.score_cut = score_cut
        self.verbose = True
        self.fit_spikes = fit_spikes

    def analyze_data(self, pred_data: List[np.ndarray], response_data: np.ndarray) -> _BaseData:
        """
        Process given data with MINE
        :param pred_data: Predictor data as a list of n_timepoints long vectors. Predictors are shared among all
            responses
        :param response_data: n_responses x n_timepoints matrix of responses
        :return:
            MineData object with the requested data
        """
        # check for matching sizes
        res_len = response_data.shape[1]
        for i, pd in enumerate(pred_data):
            if pd.size != res_len:
                raise ValueError(f"Predictor {i} has a different number of timesteps than the responses. {pd.size} vs. "
                                 f"{res_len}")
        # warn user if data is not standardized - however for spiking analysis data should not be standardized!
        if not self.fit_spikes:
            if (not np.allclose(np.mean(response_data, 1), 0, atol=1e-2)
                    or not np.allclose(np.std(response_data, 1), 1, atol=1e-2)):
                warnings.warn("WARNING: Response data does not appear standardized to 0 mean and standard deviation 1",
                              MineWarning)
        else:
            if not np.all(np.logical_or(response_data == 0 , response_data == 1)):
                warnings.warn("WARNING: Spike data analysis selected but at least part of the data is neither 1 nor 0",
                              MineWarning)
        # predictors should ideally always be standardized
        for i, pd in enumerate(pred_data):
            if not np.isclose(np.mean(pd), 0, atol=1e-2) or not np.isclose(np.std(pd), 1, atol=1e-2):
                warnings.warn(f"WARNING: Predictor {i} does not appear standardized to 0 mean and standard deviation 1",
                              MineWarning)
        train_frames = int(self.train_fraction * res_len)
        n_predictors = len(pred_data)

        # define our score function
        if self.fit_spikes:
            # the spiking model returns log-probabilities by default, hence need to transform!
            score_function = lambda predicted, real: roc_auc_score(real, utilities.sigmoid(predicted))
        else:
            score_function = lambda predicted, real: np.corrcoef(predicted, real)[0, 1]

        scores_trained = np.full(response_data.shape[0], np.nan)
        scores_test = scores_trained.copy()
        if self.compute_taylor:
            n_taylor = (n_predictors ** 2 - n_predictors) // 2 + n_predictors
            taylor_scores = np.full((response_data.shape[0], n_taylor, 2), np.nan)
            taylor_true_change = []
            taylor_full_prediction = []
            taylor_by_pred = []
            lin_approx_scores = scores_test.copy()
            me_scores = scores_test.copy()
        else:
            taylor_scores = None
            taylor_true_change = None
            taylor_full_prediction = None
            taylor_by_pred = None
            lin_approx_scores = None
            me_scores = None
        if self.return_jacobians:
            all_jacobians = np.full((response_data.shape[0], self.model_history * n_predictors), np.nan)
        else:
            all_jacobians = None
        if self.return_hessians:
            all_hessians = np.full((response_data.shape[0], self.model_history * n_predictors,
                                    self.model_history * n_predictors), np.nan)
        else:
            all_hessians = None

        data_obj = utilities.Data(self.model_history, pred_data, response_data, train_frames)
        # create model once
        m = model.get_standard_model(self.model_history, self.fit_spikes)
        # the following is required to init variables at desired shape
        m(np.random.randn(1, self.model_history, len(data_obj.regressors)).astype(np.float32))
        # save untrained weights to reinitialize model without having to recreate the class which somehow leaks memory
        init_weights = m.get_weights()
        for cell_ix in range(data_obj.ca_responses.shape[0]):
            tset = data_obj.training_data(cell_ix, batch_size=256)
            regressors = data_obj.regressor_matrix(cell_ix)
            # reset weights to pre-trained state
            m.set_weights(init_weights)
            # the following appears to be required to re-init variables?
            m(np.random.randn(1, self.model_history, len(data_obj.regressors)).astype(np.float32))
            # train
            model.train_model(m, tset, self.n_epochs, data_obj.ca_responses.shape[1])
            if self.model_weight_store is not None:
                w_group = self.model_weight_store.create_group(f"cell_{cell_ix}_weights")
                utilities.modelweights_to_hdf5(w_group, m.get_weights())
            # evaluate
            p, r = data_obj.predict_response(cell_ix, m)
            c_tr = score_function(p[:train_frames], r[:train_frames])
            scores_trained[cell_ix] = c_tr
            c_ts = score_function(p[train_frames:], r[train_frames:])
            scores_test[cell_ix] = c_ts
            # if the cell doesn't have a test score of at least score_cut we skip the rest
            # NOTE: This means that some return values will only have one entry for each unit
            # that made the cut - the user will have to handle those cases
            if c_ts < self.score_cut or not np.isfinite(c_ts):
                if self.verbose:
                    print(f"        Unit {cell_ix+1} out of {response_data.shape[0]} fit. "
                          f"Test score={scores_test[cell_ix]} which was below cut-off.")
                continue
            # compute first and second order derivatives
            tset = data_obj.training_data(cell_ix, 256)
            all_inputs = []
            for inp, outp in tset:
                all_inputs.append(inp.numpy())
            x_bar = np.mean(np.vstack(all_inputs), 0, keepdims=True)
            jacobian, hessian = d2ca_dr2(m, x_bar)
            # compute taylor-expansion and nonlinearity evaluation if requested
            if self.compute_taylor:
                # compute taylor expansion
                true_change, pc, by_pred = taylor_decompose(m, regressors, self.taylor_pred_every,
                                                            self.taylor_look_ahead)
                taylor_true_change.append(true_change)
                taylor_full_prediction.append(pc)
                taylor_by_pred.append(by_pred)
                # compute first and 2nd order model predictions
                # for spiking models these need to be computed in probability space not log-probability space
                # since deviations at the extremes in log space do not carry the same wait as deviations close to 0
                lin_score, o2_score = complexity_scores(m, x_bar, jacobian, hessian, regressors, self.taylor_pred_every,
                                                        self.fit_spikes)
                lin_approx_scores[cell_ix] = lin_score
                me_scores[cell_ix] = o2_score
                # compute our by-predictor taylor importance as the fractional loss of r2 when excluding the component
                # for spiking models these need to be computed in probability space not log-probability space
                # since deviations at the extremes in log space do not carry the same wait as deviations close to 0
                if self.fit_spikes:
                    true_change = utilities.sigmoid(true_change)
                    pc = utilities.sigmoid(pc)
                    by_pred = utilities.sigmoid(by_pred)
                off_diag_index = 0
                for row in range(n_predictors):
                    for column in range(n_predictors):
                        if row == column:
                            remainder = pc - by_pred[:, row, column]
                            # Store in the first n_diag indices of taylor_by_pred (i.e. simply at row as indexer)
                            bsample = utilities.bootstrap_fractional_r2loss(true_change, pc, remainder, 1000)
                            taylor_scores[cell_ix, row, 0] = np.mean(bsample)
                            taylor_scores[cell_ix, row, 1] = np.std(bsample)
                        elif row < column:
                            remainder = pc - by_pred[:, row, column] - by_pred[:, column, row]
                            # Store in row-major order in taylor_by_pred after the first n_diag indices
                            bsample = utilities.bootstrap_fractional_r2loss(true_change, pc, remainder, 1000)
                            taylor_scores[cell_ix, n_predictors + off_diag_index, 0] = np.mean(bsample)
                            taylor_scores[cell_ix, n_predictors + off_diag_index, 1] = np.std(bsample)
                            off_diag_index += 1
            if self.return_jacobians or self.return_hessians:
                # compute average predictor
                if self.return_jacobians:
                    jacobian = jacobian.numpy().ravel()
                    # reorder jacobian by n_predictor long chunks of hist_steps timeslices
                    jacobian = np.reshape(jacobian, (self.model_history, n_predictors)).T.ravel()
                    all_jacobians[cell_ix, :] = jacobian
                if self.return_hessians:
                    hessian = np.reshape(hessian.numpy(), (x_bar.shape[2] * self.model_history,
                                                           x_bar.shape[2] * self.model_history))
                    hessian = utilities.rearrange_hessian(hessian, n_predictors, self.model_history)
                    all_hessians[cell_ix, :, :] = hessian
            if self.verbose:
                print(f"        Unit {cell_ix+1} out of {response_data.shape[0]} completed. "
                      f"Test score={scores_test[cell_ix]}")
        if self.compute_taylor:
            # turn the taylor predictions into ndarrays unless no unit passed threshold
            if len(taylor_true_change) > 0:
                if data_obj.ca_responses.shape[0] > 1:
                    taylor_true_change = np.vstack(taylor_true_change)
                    taylor_full_prediction = np.vstack(taylor_full_prediction)
                    taylor_by_pred = np.vstack([pbp[None, :] for pbp in taylor_by_pred])
                else:
                    # only one fit object, just expand dimension to keep things consistent
                    taylor_true_change = taylor_true_change[0][None, :]
                    taylor_full_prediction = taylor_full_prediction[0][None, :]
                    taylor_by_pred = taylor_by_pred[0][None, :]
            else:
                taylor_true_change = np.nan
                taylor_full_prediction = np.nan
                taylor_by_pred = np.nan
        if self.fit_spikes:
            return_data = MineSpikingData(
                roc_auc_test=scores_test,
                roc_auc_trained=scores_trained,
                taylor_scores=taylor_scores,
                taylor_true_change=taylor_true_change,
                taylor_full_prediction=taylor_full_prediction,
                taylor_by_predictor=taylor_by_pred,
                model_lin_approx_scores=lin_approx_scores,
                model_2nd_approx_scores=me_scores,
                jacobians=all_jacobians,
                hessians=all_hessians
            )
        else:
            return_data = MineData(
                correlations_test=scores_test,
                correlations_trained=scores_trained,
                taylor_scores=taylor_scores,
                taylor_true_change=taylor_true_change,
                taylor_full_prediction=taylor_full_prediction,
                taylor_by_predictor=taylor_by_pred,
                model_lin_approx_scores=lin_approx_scores,
                model_2nd_approx_scores=me_scores,
                jacobians=all_jacobians,
                hessians=all_hessians
            )
        return return_data



"""
Module for all network models with tensorflow dependency
"""
import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras import layers, regularizers, initializers
from typing import Optional, Union

class NotInitialized(Exception):
    def __init__(self, message):
        super().__init__(message)


class ActivityPredictor(keras.Model):
    """
    Simple network for non-linear prediction of calcium activity
    """

    def __init__(self, n_units: int, n_conv: int, drop_rate: float, input_length: int, activation: str,
                 predict_spikes: bool):
        """
        Creates a new NetNavigator
        :param n_units: The number of units in each dense layer
        :param n_conv: The number of units in each initial convolutional layer
        :param drop_rate: The drop-out rate during training
        :param input_length: The length (across time) of inputs to the network (sets conv filter size)
        :param activation: The activation function to use
        :param predict_spikes: If true, 0/1 spike data instead of continuous data is expected
        """
        super(ActivityPredictor, self).__init__()
        if drop_rate < 0 or drop_rate > 1:
            raise ValueError("drop_rate has to be between 0 and 1")
        if n_units < 1:
            raise ValueError("Need at least one unit in each dense layer")
        if n_conv < 1:
            raise ValueError("Need at least one convolutional unit")
        self._n_units: int = n_units
        self._n_conv: int = n_conv
        self.input_length: int = input_length
        self._activation: str = activation
        self._drop_rate: float = drop_rate
        self.l2_sparsity: float = 2e-4  # sparsity constraint on weight vectors
        self.learning_rate: float = 1e-3
        # optimizer and loss functions
        self.optimizer: Optional[keras.optimizers.Optimizer] = None
        self.loss_fn: Optional[keras.losses.Loss] = None
        self._initialized: bool = False
        # output placeholders
        self._out: Optional[keras.layers.Layer] = None
        # layers
        self._conv_layer: Optional[keras.layers.Layer] = None  # Convolutional layer
        self._drop_cl: Optional[keras.layers.Layer] = None  # Dropout of convolutional layer
        self._deep_1: Optional[keras.layers.Layer] = None  # First deep layer
        self._drop_d1: Optional[keras.layers.Layer] = None  # Dropout of first deep layer
        self._deep_2: Optional[keras.layers.Layer] = None  # Second deep layer
        self._drop_d2: Optional[keras.layers.Layer] = None  # Dropout of second deep layer

        self._deep_3: Optional[keras.layers.Layer] = None  # Third deep layer
        self._drop_d3: Optional[keras.layers.Layer] = None  # Dropout of third deep layer
        self._deep_4: Optional[keras.layers.Layer] = None  # Fourth deep layer
        self._drop_d4: Optional[keras.layers.Layer] = None  # Dropout of fourth deep layer
        self._flatten: Optional[keras.layers.Layer] = None
        # model-specific cash field used during derivative calculation
        self.part_tensor_1: Optional[np.ndarray] = None
        # Store if this is a model to predict continuous data or spikes
        self.predict_spikes = predict_spikes

    def setup(self) -> None:
        """
        Initializes the model, resetting weights
        """
        # processing
        self._conv_layer = layers.Conv1D(filters=self.n_conv,
                                         kernel_size=self.input_length,
                                         # set kernel size = input size => computes dot product
                                         use_bias=True,
                                         padding='valid',
                                         activation=self.activation,
                                         kernel_initializer=initializers.GlorotUniform(),
                                         kernel_regularizer=regularizers.l2(self.l2_sparsity),
                                         strides=1, name="Convolution")
        self._flatten = layers.Flatten()
        self._drop_cl = layers.Dropout(self.drop_rate)
        self._deep_1 = layers.Dense(units=self.n_units, activation=self.activation,
                                    kernel_initializer=initializers.GlorotUniform(),
                                    kernel_regularizer=regularizers.l2(self.l2_sparsity), name="Deep1")
        self._drop_d1 = layers.Dropout(self.drop_rate)
        self._deep_2 = layers.Dense(units=self.n_units, activation=self.activation,
                                    kernel_initializer=initializers.GlorotUniform(),
                                    kernel_regularizer=regularizers.l2(self.l2_sparsity), name="Deep2")
        self._drop_d2 = layers.Dropout(self.drop_rate)

        self._deep_3 = layers.Dense(units=self.n_units, activation=self.activation,
                                    kernel_initializer=initializers.GlorotUniform(),
                                    kernel_regularizer=regularizers.l2(self.l2_sparsity), name="Deep3")
        self._drop_d3 = layers.Dropout(self.drop_rate)
        self._deep_4 = layers.Dense(units=self.n_units, activation=self.activation,
                                    kernel_initializer=initializers.GlorotUniform(),
                                    kernel_regularizer=regularizers.l2(self.l2_sparsity), name="Deep4")
        self._drop_d4 = layers.Dropout(self.drop_rate)
        # output: This is just one value, that should predict the calcium response at the current time
        self._out = layers.Dense(1, activation=None, name="Out")
        # create our optimizer and loss functions
        self.optimizer = keras.optimizers.legacy.Adam(self.learning_rate)
        # all we need to do to be able to predict spikes is change our loss function! This change is transparent
        if self.predict_spikes:
            self.loss_fn = keras.losses.BinaryCrossentropy(from_logits=True)
        else:
            self.loss_fn = keras.losses.MeanSquaredError()
        self._initialized = True

    def get_output(self, inputs: np.ndarray) -> float:
        """
        Returns the output value given the model inputs
        :param inputs: batchsize x input_length x n_regressors (the channels)
        :return: 1 output value corresponding to the calcium response
        """
        self.check_input(inputs)
        out = self(inputs)
        return out.numpy().ravel()

    def get_probability(self, inputs: np.ndarray) -> float:
        """
        Returns the spike probability given model inputs
        :param inputs: batchsize x input_length x n_regressors (the channels)
        :return: 1 output value corresponding to the spike probability
        """
        if not self.predict_spikes:
            raise ValueError("Model does not predict spikes. Probability representation is meaningless")
        self.check_input(inputs)
        logit_out = self(inputs)
        return tf.math.sigmoid(logit_out).numpy().ravel()

    def clear_model(self) -> None:
        """
        Clears and uninitializes the model
        """
        self._conv_layer = None  # Convolutional layer
        self._drop_cl = None  # Dropout of convolutional layer
        self._deep_1 = None  # First deep layer
        self._drop_d1 = None  # Dropout of first deep layer
        self._deep_2 = None  # Second deep layer
        self._drop_d2 = None  # Dropout of second deep layer

        self._deep_3 = None  # Third deep layer
        self._drop_d3 = None  # Dropout of third deep layer
        self._deep_4 = None  # Fourth deep layer
        self._drop_d4 = None  # Dropout of fourth deep layer
        self._out = None
        self._initialized = False

    def check_init(self) -> None:
        if not self._initialized:
            raise NotInitialized("Model not initialized. Call setup or load.")

    def check_input(self, inputs) -> None:
        if inputs.shape[1] != self.input_length:
            raise ValueError("Input length across time different than expected")

    @tf.function
    def call(self, inputs: Union[np.ndarray, tf.Tensor], training: Optional[bool] = None, mask=None) -> tf.Tensor:
        if training is None:
            training = False
        self.check_init()
        inputs = self._conv_layer(inputs, training=training)
        inputs = self._flatten(inputs)
        inputs = self._drop_cl(inputs, training=training)
        inputs = self._deep_1(inputs, training=training)
        inputs = self._drop_d1(inputs, training=training)
        inputs = self._deep_2(inputs, training=training)
        inputs = self._drop_d2(inputs, training=training)

        inputs = self._deep_3(inputs, training=training)
        inputs = self._drop_d3(inputs, training=training)
        inputs = self._deep_4(inputs, training=training)
        inputs = self._drop_d4(inputs, training=training)
        return self._out(inputs)

    @tf.function
    def train_step(self, btch_inputs: Union[np.ndarray, tf.Tensor], btch_labels: Union[np.ndarray, tf.Tensor]):
        """
        Runs one training step on the model
        :param btch_inputs: The inputs of the batch
        :param btch_labels: True values of the calcium response
        """
        with tf.GradientTape() as tape:
            pred = self(btch_inputs, training=True)
            loss = self.loss_fn(btch_labels, pred)
            # add any extra losses created during foward pass
            # NOTE: I'm not entirely clear if this is really required when the only extra losses
            # are kernel regularizations, however the values match the absolute weights exactly
            # i.e. they are the L1 loss
            loss += sum(self.losses)
        gradients = tape.gradient(loss, self.trainable_variables)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))

    @property
    def activation(self) -> str:
        return self._activation

    @property
    def drop_rate(self) -> float:
        return self._drop_rate

    @property
    def n_units(self) -> int:
        return self._n_units

    @property
    def n_conv(self) -> int:
        return self._n_conv

    @property
    def conv_layer(self) -> Optional[keras.layers.Layer]:
        return self._conv_layer

    @property
    def deep_1(self) -> Optional[keras.layers.Layer]:
        return self._deep_1

    @property
    def deep_2(self) -> Optional[keras.layers.Layer]:
        return self._deep_2

    @property
    def deep_3(self) -> Optional[keras.layers.Layer]:
        return self._deep_3

    @property
    def deep_4(self) -> Optional[keras.layers.Layer]:
        return self._deep_4

    @property
    def flatten(self) -> Optional[keras.layers.Layer]:
        return self._flatten

    @property
    def out(self) -> Optional[keras.layers.Layer]:
        return self._out


def train_model(mdl: ActivityPredictor, tset: tf.data.Dataset, n_epochs: int, datacount: int) -> None:
    """
    Trains the model over n epochs. NOTE: The dataset passed should be generated from a shuffle with
    reshuffle_each_iteration set to true to ensure that data is presented in shuffled order in each
    training epoch
    :param mdl: The model to train
    :param tset: The dataset with training data
    :param n_epochs: The number of epochs to train for
    :param datacount: Unused
    """
    for e in range(n_epochs):
        for inp, outp in tset:
            mdl.train_step(inp, outp)


def get_standard_model(hist_steps: int, predict_spikes: bool) -> ActivityPredictor:
    """
    Creates and returns an activity predictor instance with standard parameters
    found through a hyperparameter search
    :param hist_steps: The number of history steps in the model
    :param predict_spikes: If true, 0/1 spike data instead of continuous data is expected
    """
    m = ActivityPredictor(64, 80, 0.5, hist_steps, "swish", predict_spikes)
    m.learning_rate = 1e-3
    m.l2_sparsity = 1e-5
    m.setup()
    return m



"""
Script to convert Suite2p extracted activity into CSV format for MINE ingestion
"""

from PyQt5.QtWidgets import QFileDialog, QApplication
import numpy as np
import pandas as pd
import argparse
from os import path
import os


if __name__ == '__main__':
    app = QApplication([])
    a_parser = argparse.ArgumentParser(prog="convert_suite2p.py",
                                       description="Converts suite2p output data to CSV format for MINE ingestion.")
    a_parser.add_argument("-d", "--s2pdir", help="Path to suite2p directory.", type=str)
    a_parser.add_argument("-isc", "--th_iscell", help="Threshold on suite2p classifier to consider "
                                                      "object as a neuron.", type=float, default=0.5)
    a_parser.add_argument("-ft", "--frame_time", help="Time between consecutive frames in same unit "
                                                      "as used for predictors. If not set suit2p value will be used.",
                          type=float)
    a_parser.add_argument("-dc", "--deconv", help="If set use deconvolved data instead of fluorescence.",
                          action='store_true')

    args = a_parser.parse_args()
    th_iscell = args.th_iscell
    frame_time = args.frame_time
    deconv = args.deconv
    s2pdir = args.s2pdir
    if s2pdir is None:
        s2pdir = QFileDialog.getExistingDirectory(caption="Please select suite2p output directory",
                                                  options=QFileDialog.DontUseNativeDialog)

    plane_dirs = os.listdir(s2pdir)
    plane_dirs = [path.join(s2pdir, d) for d in plane_dirs if path.isdir(path.join(s2pdir, d))] + [s2pdir]
    val_dirs = [d for d in plane_dirs if path.exists(path.join(d, "ops.npy"))]

    if len(val_dirs) == 0:
        app.quit()
        del app
        raise ValueError("Directory does not seem to be valid suite2p output."
                         "No subdirectories with output files found.")

    for vd in val_dirs:

        if frame_time is None:
            ops = np.load(path.join(vd, "ops.npy"), allow_pickle=True)
            fs = ops[()]["fs"]  # this is a frame-rate in Hz!
            ft = 1 / fs
        else:
            ft = frame_time

        # Suite2p saves the neural data such that neurons are index along axis 0
        # time is along axis 1 - we therefore need to transpose before saving
        if deconv:
            neuro_data = np.load(path.join(vd, "spks.npy"))
        else:
            neuro_data = np.load(path.join(vd, "F.npy"))

        iscell = np.load(path.join(vd, "iscell.npy"))[:, 1] > th_iscell

        neuro_data = neuro_data[iscell]

        time = np.arange(neuro_data.shape[1])*ft

        full_data = np.c_[time[:, None], neuro_data.T]

        col_names = ["time"] + [f"Neuron {i}" for i in range(neuro_data.shape[0])]

        full_data = pd.DataFrame(data=full_data, columns=col_names)

        full_data.to_csv(path.join(vd, "Neuron_data.csv"), index=False)

    app.quit()
    del app



import datetime
import importlib.resources
import json
from PySide6.QtCore import QProcess
from PySide6.QtGui import QPalette, QColor
from PySide6.QtWidgets import QApplication, QWidget, QFileDialog, QLineEdit, QCheckBox, QMessageBox
from neuro_mine.ui.ui_form import Ui_Widget
import numpy as np
import os
from process_csv import default_options
import subprocess
import sys

class MyApp(QWidget, Ui_Widget):
    def __init__(self):
        super().__init__()
        self.setupUi(self)
        self.default_options = default_options

        now = datetime.datetime.now().strftime("%B_%d_%Y_%I_%M%p")
        self.lineEdit.setText(now) # Model Name
        self.checkBox.setChecked(default_options["use_time"]) # Use Time as Predictor
        self.checkBox_2.setChecked(default_options["run_shuffle"]) # Shuffle Data
        self.lineEdit_3.setText(str(default_options["th_test"])) # Test Score Threshold
        self.lineEdit_5.setText(str(default_options["taylor_sig"])) # Taylor Expansion Significance Threshold
        self.lineEdit_11.setText(str(default_options["taylor_look"]))  # Taylor Cutoff
        self.lineEdit_6.setText(str(default_options["taylor_cut"])) # Taylor Cutoff
        self.lineEdit_7.setText(str(default_options["th_lax"]))  # Linear Fit Variance explained cutoff
        self.lineEdit_8.setText(str(default_options["th_sqr"])) # Square Fit Variance explained cutoff
        self.checkBox_4.setChecked(default_options["jacobian"]) # Store Linear Receptive Fields (Jacobians)
        self.lineEdit_9.setText(str(default_options["history"])) # Model History [s]
        self.lineEdit_12.setText(str(default_options["n_epochs"])) # Number of Epochs
        self.lineEdit_13.setText(str(default_options["miner_train_fraction"])) # Fraction of Data to use to Train
        self.checkBox_3.setChecked(True) # Verbose Fitting Updates

        # connect signals
        self.pushButton.clicked.connect(self.on_run_clicked)
        self.pushButton_2.clicked.connect(lambda: self.browse_file(self.lineEdit_4, "Predictor", "*.csv"))
        self.pushButton_3.clicked.connect(lambda: self.browse_file(self.lineEdit_2, "Response", "*.csv"))
        self.pushButton_4.clicked.connect(lambda: self.handle_json_browse(self.lineEdit_10))
        self.pushButton_5.clicked.connect(self.restore_defaults)
        self.pushButton_6.clicked.connect(self.save_to_json)

        # connect field validation
        self.valid_fields = {}
        for le, minv, maxv in [
            (self.lineEdit_3, 0, 1),
            (self.lineEdit_5, 0, 1),
            (self.lineEdit_11, 0.00000001, 3.999999999),
            (self.lineEdit_6, 0, 1),
            (self.lineEdit_7, 0, 1),
            (self.lineEdit_8, 0, 1),
            (self.lineEdit_9, 1.0, np.inf),
            (self.lineEdit_12, 0, 100),
            (self.lineEdit_13, 0, 1)
        ]:
            le.editingFinished.connect(lambda le=le, minv=minv, maxv=maxv: self.validate_range(le, minv, maxv))

        self.last_dir = ""

        self.lineEdit_4.textChanged.connect(self.update_button_states)
        self.lineEdit_2.textChanged.connect(self.update_button_states)

        self.update_button_states()

    def browse_file(self, target_lineedit, file_type, file_filter):
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            f"Select {file_type}",
            self.last_dir or "",
            file_filter,
            options=QFileDialog.DontUseNativeDialog
        )
        if file_path:
            target_lineedit.setText(file_path)

    def populate_presets(self):
        for attr, value in default_options["line_edits"].items():
            widget = getattr(self, attr, None)
            if isinstance(widget, QLineEdit):
                widget.setText(value)

        for attr, value in default_options["check_boxes"].items():
            widget = getattr(self, attr, None)
            if isinstance(widget, QCheckBox):
                widget.setChecked(value)

    def save_to_json(self):
        data = {
            "config": {
                "use_time":self.checkBox.isChecked(),
                "run_shuffle":self.checkBox_2.isChecked(),
                "th_test":self.lineEdit_3.text().strip(),
                "taylor_sig":self.lineEdit_5.text().strip(),
                "taylor_cut":self.lineEdit_6.text().strip(),
                "th_lax":self.lineEdit_7.text().strip(),
                "th_sqr":self.lineEdit_8.text().strip(),
                "history":self.lineEdit_9.text().strip(),
                "taylor_look":self.lineEdit_11.text().strip(),
                "jacobian":self.checkBox_4.isChecked(),
                "n_epochs":self.lineEdit_12.text().strip(),
                "miner_verbose":self.checkBox_3.isChecked(),
                "miner_train_fraction":self.lineEdit_13.text().strip()
                }
        }

        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save Configuration",
            self.last_dir or "",
            "JSON Files (*.json);;All Files (*)",
            options=QFileDialog.DontUseNativeDialog
        )

        if file_path:
            if not file_path.lower().endswith(".json"):
                file_path += ".json"

            try:
                with open(file_path, "w", encoding="utf-8") as f:
                    json.dump(data, f, indent=4)
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Failed to save file:\n{e}")

            self.last_dir = os.path.dirname(file_path)

    def handle_json_browse(self, target_lineedit):
        file_path, _ = QFileDialog.getOpenFileName(self, "Select JSON File", "", "JSON Files (*.json)")
        if file_path:
            target_lineedit.setText(file_path)
            self.load_json_and_populate(file_path)

    def load_json_and_populate(self, file_path):
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                data = json.load(f)
            if "config" in data:
                data = data["config"]
        except Exception as e:
            QMessageBox.critical(self, "Error", f"Could not read JSON:\n{e}")
            return

        self.checkBox.setChecked(data.get("use_time", default_options["use_time"]))
        self.checkBox_2.setChecked(data.get("run_shuffle", default_options["run_shuffle"]))
        self.lineEdit_3.setText(str(data.get("th_test", default_options["th_test"])))
        self.lineEdit_5.setText(str(data.get("taylor_sig", default_options["taylor_sig"])))
        self.lineEdit_6.setText(str(data.get("taylor_cut", default_options["taylor_cut"])))
        self.lineEdit_7.setText(str(data.get("th_lax", default_options["th_lax"])))
        self.lineEdit_8.setText(str(data.get("th_sqr", default_options["th_sqr"])))
        self.lineEdit_9.setText(str(data.get("history", default_options["history"])))
        self.lineEdit_11.setText(str(data.get("taylor_look", default_options["taylor_look"])))
        self.checkBox_4.setChecked(data.get("jacobian", default_options["jacobian"]))
        self.lineEdit_12.setText(str(data.get("n_epochs", default_options["n_epochs"])))
        self.checkBox_3.setChecked(data.get("miner_verbose", default_options["miner_verbose"]))
        self.lineEdit_13.setText(str(data.get("miner_train_fraction", default_options["miner_train_fraction"])))

    def validate_range(self, line_edit, min_val, max_val):
        text = line_edit.text().strip()

        try:
            value = float(text)
            if min_val <= value <= max_val:
                line_edit.setPalette(self.style().standardPalette())
                self.valid_fields[line_edit.objectName()] = True
            else:
                palette = line_edit.palette()
                palette.setColor(QPalette.Base, QColor("crimson"))
                line_edit.setPalette(palette)
                self.valid_fields[line_edit.objectName()] = False
        except ValueError:
            palette = line_edit.palette()
            palette.setColor(QPalette.Base, QColor("crimson"))
            line_edit.setPalette(palette)
            self.valid_fields[line_edit.objectName()] = False

        self.update_button_states()

    def update_button_states(self):
        all_valid = all(self.valid_fields.values())

        line4_filled = bool(self.lineEdit_4.text().strip())
        line2_filled = bool(self.lineEdit_2.text().strip())
        required_fields_filled = line4_filled and line2_filled

        self.pushButton.setEnabled(all_valid and required_fields_filled)

        self.pushButton_6.setEnabled(all_valid)

    def restore_defaults(self):
        """Restore UI elements to their default preset values."""
        global default_options

        self.checkBox.setChecked(default_options["use_time"])
        self.checkBox_2.setChecked(default_options["run_shuffle"])
        self.lineEdit_3.setText(str(default_options["th_test"]))
        self.lineEdit_5.setText(str(default_options["taylor_sig"]))
        self.lineEdit_11.setText(str(default_options["taylor_look"]))
        self.lineEdit_6.setText(str(default_options["taylor_cut"]))
        self.lineEdit_7.setText(str(default_options["th_lax"]))
        self.lineEdit_8.setText(str(default_options["th_sqr"]))
        self.checkBox_4.setChecked(default_options["jacobian"])
        self.lineEdit_9.setText(str(default_options["history"]))
        self.lineEdit_12.setText(str(default_options["n_epochs"]))
        self.lineEdit_13.setText(str(default_options["miner_train_fraction"]))
        self.checkBox_3.setChecked(True)

        self.reset_validation_state()

    def reset_validation_state(self):
        """Resets line edit colors and re-enables buttons after restoring defaults."""
        for widget in [self.lineEdit_3, self.lineEdit_5, self.lineEdit_11,
                       self.lineEdit_6, self.lineEdit_7, self.lineEdit_8,
                       self.lineEdit_9, self.lineEdit_12, self.lineEdit_13]:
            widget.setPalette(self.style().standardPalette())

        for le, minv, maxv in [
            (self.lineEdit_3, 0, 1),
            (self.lineEdit_5, 0, 1),
            (self.lineEdit_11, 0.00000001, 3.999999999),
            (self.lineEdit_6, 0, 1),
            (self.lineEdit_7, 0, 1),
            (self.lineEdit_8, 0, 1),
            (self.lineEdit_9, 1.0, np.inf),
            (self.lineEdit_12, 0, 100),
            (self.lineEdit_13, 0, 1)
        ]:
            le.editingFinished.connect(lambda le=le, minv=minv, maxv=maxv: self.validate_range(le, minv, maxv))

    def on_run_clicked(self):

        model_name = self.lineEdit.text()
        predictors = self.lineEdit_4.text()
        responses = self.lineEdit_2.text()
        use_time = self.checkBox.isChecked()
        run_shuffle = self.checkBox_2.isChecked()
        th_test = self.lineEdit_3.text()
        taylor_sig = self.lineEdit_5.text()
        taylor_cut = self.lineEdit_6.text()
        th_lax = self.lineEdit_7.text()
        th_sqr = self.lineEdit_8.text()
        history = self.lineEdit_9.text()
        taylor_look = self.lineEdit_11.text()
        jacobian = self.checkBox_4.isChecked()
        config = self.lineEdit_10.text()
        n_epochs = self.lineEdit_12.text()
        miner_verbose = self.checkBox_3.isChecked()
        miner_train_fraction = self.lineEdit_13.text()

        with importlib.resources.path("neuro_mine.scripts", "process_csv.py") as script_path:
            args = [sys.executable, str(script_path)]

            if model_name:
                args.extend(["--model_name", model_name])
            if predictors:
                args.extend(["--predictors", predictors])
            if responses:
                args.extend(["--responses", responses])
            if use_time:
                args.extend(["--use_time"])
            if run_shuffle:
                args.extend(["--run_shuffle"])
            if th_test:
                args.extend(["--th_test", th_test])
            if taylor_sig:
                args.extend(["--taylor_sig", taylor_sig])
            if taylor_cut:
                args.extend(["--taylor_cut", taylor_cut])
            if th_lax:
                args.extend(["--th_lax", th_lax])
            if th_sqr:
                args.extend(["--th_sqr", th_sqr])
            if history:
                args.extend(["--history", history])
            if taylor_look:
                args.extend(["--taylor_look", taylor_look])
            if jacobian:
                args.extend(["--jacobian"])
            if config:
                args.extend(["--config", config])
            if n_epochs:
                args.extend(["--n_epochs", n_epochs])
            if miner_verbose:
                args.extend(["--miner_verbose"])
            if miner_train_fraction:
                args.extend(["--miner_train_fraction", miner_train_fraction])

            subprocess.run(args)

        self.pushButton.setText("Running Model...")

        QApplication.quit()

def run_ui():
    app = QApplication(sys.argv)
    window = MyApp()
    window.show()
    app.exec()

if __name__ == "__main__":
    run_ui()



# This Python file uses the following encoding: utf-8
import sys

from PySide6.QtWidgets import QApplication, QWidget

# Important:
# You need to run the following command to generate the ui_form.py file
#     pyside6-uic form.ui -o ui_form.py, or
#     pyside2-uic form.ui -o ui_form.py
from ui_form import Ui_Widget

class Widget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_Widget()
        self.ui.setupUi(self)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    widget = Widget()
    widget.show()
    sys.exit(app.exec())



from typing import Dict
import numpy as np
import pandas as pd
from os import path
import os
import h5py
import file_handling as fh
import json
from utilities import safe_standardize, interp_events
from mine import Mine
import upsetplot as ups
import matplotlib.pyplot as pl


def process_file_pair(resp_path: str, pred_path: str, configuration: Dict):
    your_model = configuration["run"]["model_name"]
    run_shuffle = configuration["config"]["run_shuffle"]
    time_as_pred = configuration["config"]["use_time"]
    history_time = configuration["config"]["history"]
    taylor_look_fraction = configuration["config"]["taylor_look"]
    miner_train_fraction = configuration["config"]["miner_train_fraction"]
    test_score_thresh = configuration["config"]["th_test"]
    fit_jacobian = configuration["config"]["jacobian"]
    fit_epochs = configuration["config"]["n_epochs"]
    miner_verbose = configuration["config"]["miner_verbose"]
    taylor_sig = configuration["config"]["taylor_sig"]
    lax_thresh = configuration["config"]["th_lax"]
    sqr_thresh = configuration["config"]["th_sqr"]
    taylor_cutoff = configuration["config"]["taylor_cut"]

    resp_data, resp_has_header, resp_header = fh.CSVParser(resp_path, "R").load_data()
    pred_data, pred_has_header, pred_header = fh.CSVParser(pred_path, "P").load_data()

    # store all output file in a sub-folder of the response file folder
    output_folder = path.join(path.split(resp_path)[0], "output")
    if not path.exists(output_folder):
        os.makedirs(output_folder)

    # We use a very simple heuristic to detect spiking data and we will not allow for mixed data. In other words
    # a response file either contains all continuous data or all spiking data. When in doubt, we will treat as
    # continuous
    if np.all(np.logical_or(resp_data==0, resp_data==1)):
        is_spike_data = True
        print("Responses are assumed to contain spikes")
    else:
        is_spike_data = False
        print("Responses are assumed to be continuous values not spikes")

    pred_times = pred_data[:, 0]
    resp_times = resp_data[:, 0]

    # define interpolation time as the timespan covered in both files at the rate in the file with fewer timepoints
    # within that timespan (i.e. we bin to the lower resolution instead of interpolating to the higher resolution)
    max_allowed_time = min([pred_times.max(), resp_times.max()])
    min_allowed_time = max([pred_times.min(), resp_times.min()])
    valid_pred = np.logical_and(pred_times <= max_allowed_time, pred_times >= min_allowed_time)
    valid_resp = np.logical_and(resp_times <= max_allowed_time, resp_times >= min_allowed_time)
    # define interpolation time based on the less dense data ensuring equal timesteps
    if np.sum(valid_pred) < np.sum(valid_resp):
        ip_time = np.linspace(min_allowed_time, max_allowed_time, np.sum(valid_pred))
    else:
        ip_time = np.linspace(min_allowed_time, max_allowed_time, np.sum(valid_resp))

    # perform interpolation
    ip_pred_data = np.hstack(
        [np.interp(ip_time, pred_times[valid_pred], pd[valid_pred])[:, None] for pd in pred_data.T])
    if not is_spike_data:
        ip_resp_data = np.hstack(
            [np.interp(ip_time, resp_times[valid_resp], rd[valid_resp])[:, None] for rd in resp_data.T])
    else:
        ip_resp_data = np.hstack(
            [interp_events(ip_time, resp_times[valid_resp], rd[valid_resp])[:, None] for rd in resp_data.T])

    # Save interpolated data with chosen column names
    df_ip_resp_data = pd.DataFrame(ip_resp_data, columns=resp_header)
    df_ip_resp_data.to_csv(path.join(output_folder, f"MINE_{your_model}_interpolated_responses.csv"), index=False)
    df_ip_pred_data = pd.DataFrame(ip_pred_data, columns=pred_header)
    df_ip_pred_data.to_csv(path.join(output_folder, f"MINE_{your_model}_interpolated_predictors.csv"), index=False)

    # perform data-appropriate standardization of predictors and responses
    if time_as_pred == "Y":
        mine_pred = [safe_standardize(ipd) for ipd in ip_pred_data.T]
    else:
        mine_pred = [safe_standardize(ipd) for ipd in ip_pred_data.T[1:]]
    # In the following the first column is removed since it is time
    if not is_spike_data:
        mine_resp = safe_standardize(ip_resp_data[:, 1:]).T
    else:
        mine_resp = ip_resp_data[:, 1:].T

    configuration["run"]["interpolation_time_delta"] = np.mean(np.diff(ip_time))
    configuration["run"]["is_spike_data"] = is_spike_data
    configuration["run"]["n_predictors"] = len(mine_pred)
    with open(path.join(output_folder, f"MINE_{your_model}_run_config.json"), 'w') as config_file:
        json.dump(configuration, config_file, indent=2)

    # compute our "frame rate", i.e. frames per time-unit on the interpolated scale
    ip_rate = 1 / np.mean(np.diff(ip_time))
    # based on the rate, compute the number of frames within the model history and taylor-look-ahead
    model_history = int(np.round(history_time * ip_rate, 0))
    if model_history < 1:
        model_history = 1
    taylor_look_ahead = int(np.round(model_history * taylor_look_fraction, 0))
    if taylor_look_ahead < 1:
        taylor_look_ahead = 1
    print(f"Model history is {model_history} frames")
    print(f"Taylor look ahead is {taylor_look_ahead} frames")

    ###
    # Fit model
    ###
    mdata_shuff = None

    weight_file_name = f"MINE_{your_model}_weights.hdf5"
    with h5py.File(path.join(output_folder, weight_file_name), "w") as weight_file:
        w_grp = weight_file.create_group("fit")
        miner = Mine(miner_train_fraction, model_history, test_score_thresh, True, fit_jacobian,
                     taylor_look_ahead, 5, fit_spikes=is_spike_data)
        miner.n_epochs = fit_epochs
        miner.verbose = miner_verbose
        miner.model_weight_store = w_grp
        mdata = miner.analyze_data(mine_pred, mine_resp)
        # save neuron names
        name_grp = weight_file.create_group("response_names")
        for i, r in enumerate(resp_header[1:]):  # first entry is "Time"
            name_grp.create_dataset(f"{i}", data=r.encode('utf-8'))

    # rotate mine_resp on user request and re-fit without computing any Taylor just to get test correlations
    if run_shuffle:
        mine_resp_shuff = np.roll(mine_resp, mine_resp.shape[1] // 2, axis=1)
        with h5py.File(path.join(output_folder, weight_file_name), "a") as weight_file:
            w_grp = weight_file.create_group("fit_shuffled")
            miner = Mine(miner_train_fraction, model_history, test_score_thresh, False, False,
                         taylor_look_ahead, 5, fit_spikes=is_spike_data)
            miner.n_epochs = fit_epochs
            miner.verbose = miner_verbose
            miner.model_weight_store = w_grp
            mdata_shuff = miner.analyze_data(mine_pred, mine_resp_shuff)

    full_ana_file_name = f"MINE_{your_model}_analysis.hdf5"
    with h5py.File(path.join(output_folder, full_ana_file_name), "w") as ana_file:
        ana_grp = ana_file.create_group("analysis")
        mdata.save_to_hdf5(ana_grp)
        if mdata_shuff is not None:
            ana_grp = ana_file.create_group("analysis_shuffled")
            mdata_shuff.save_to_hdf5(ana_grp)

    ###
    # Output model insights as csv
    ###
    model_scores = mdata.roc_auc_test if is_spike_data else mdata.correlations_test
    predictor_columns = pred_header if time_as_pred == 'Y' else pred_header[1:]
    interpret_dict = {"Response": [], "Fit": []} | {ph: [] for ph in predictor_columns} | {"Linearity": []}
    interpret_name = f"MINE_{your_model}_Insights.csv"
    n_objects = model_scores.size
    # for taylor analysis (which predictors are important) compute our significance levels based on a) user input
    # and b) the number of responses above threshold which gives the multiple-comparison correction - bonferroni
    min_significance = 1 - taylor_sig / np.sum(model_scores >= test_score_thresh)
    normal_quantiles_by_sigma = np.array([0.682689492137, 0.954499736104, 0.997300203937, 0.999936657516,
                                          0.999999426697, 0.999999998027])
    n_sigma = np.where((min_significance - normal_quantiles_by_sigma) < 0)[0][0] + 1

    for j in range(n_objects):
        response = resp_header[j + 1]  # because resp_header still contains the first "time" column
        interpret_dict["Response"].append(response)
        fit = model_scores[j] > test_score_thresh
        interpret_dict["Fit"].append("Y" if fit else "N")
        if not fit:
            for pc in predictor_columns:
                interpret_dict[pc].append("-")
            interpret_dict["Linearity"].append("-")
        else:
            if mdata.model_lin_approx_scores[j] >= lax_thresh:
                interpret_dict["Linearity"].append("linear")
            else:
                if mdata.model_2nd_approx_scores[j] >= sqr_thresh:
                    interpret_dict["Linearity"].append("quadratic")
                else:
                    interpret_dict["Linearity"].append("cubic+")
            for k, pc in enumerate(predictor_columns):
                taylor_mean = mdata.taylor_scores[j][k][0]
                taylor_std = mdata.taylor_scores[j][k][1]
                taylor_is_sig = taylor_mean - n_sigma * taylor_std - taylor_cutoff
                interpret_dict[pc].append("Y" if taylor_is_sig > 0 else "N")
    interpret_df = pd.DataFrame(interpret_dict)
    interpret_df.to_csv(path.join(output_folder, interpret_name), index=False)

    # save Jacobians: One CSV file for each predictor, containing the Jacobians for each response
    # column headers will be the time delay relative to t=0, since our modeling is set up
    # such that convolutions are restricted to the past (hence model_history)
    def time_from_index(ix: int) -> float:
        ix_corr = ix - model_history + 1  # at model history is timepoint 0
        return ip_rate * ix_corr

    if fit_jacobian and np.any(model_scores >= test_score_thresh):
        for i, pc in enumerate(predictor_columns):
            jac_dict = {"Response": []} | {f"{time_from_index(t)}": [] for t in range(model_history)}
            jac_file_name = f"MINE_{your_model}_ReceptiveFields_{pc}.csv"
            for j in range(n_objects):
                if np.any(np.isnan(mdata.jacobians[j, :])):
                    continue
                response = resp_header[j + 1]  # because resp_header still contains the first "time" column
                jac_dict["Response"].append(response)
                # index out the predictor related receptive field
                rf = mdata.jacobians[j, i*model_history:(i+1)*model_history]
                for t in range(model_history):
                    jac_dict[f"{time_from_index(t)}"].append(rf[t])
            df_jac = pd.DataFrame(jac_dict)
            df_jac.to_csv(path.join(output_folder, jac_file_name), index=False)

    # if shuffles were calculated plot fraction of above threshold units in data and shuffle
    # versus correlation threshold levels
    if run_shuffle:
        shuffle_scores = mdata_shuff.roc_auc_test if is_spike_data else mdata_shuff.correlations_test
        fig, axes = pl.subplots(nrows=2)
        c_thresholds = np.linspace(0, 1)
        ab_real = np.full_like(c_thresholds, np.nan)
        ab_shuff = np.full_like(c_thresholds, np.nan)
        for i, ct in enumerate(c_thresholds):
            ab_real[i] = np.sum(model_scores > ct) / n_objects
            ab_shuff[i] = np.sum(shuffle_scores > ct) / n_objects
        enrichment = ab_real / ab_shuff
        axes[0].plot(c_thresholds, ab_real, label="Real data")
        axes[0].plot(c_thresholds, ab_shuff, label="Shuffled data")
        axes[0].plot([test_score_thresh, test_score_thresh], [0, 1], 'k--', label="Threshold")
        metric_label = "ROC AUC" if is_spike_data else "Correlation"
        axes[0].set_xlabel(f"Test {metric_label} cutoff")
        axes[0].set_ylabel("Fraction above threshold")
        axes[0].set_ylim(0, 1)
        axes[0].set_xlim(0, 1)
        axes[0].legend()
        axes[1].plot(c_thresholds, enrichment)
        axes[1].plot([test_score_thresh, test_score_thresh], [np.nanmin(enrichment), np.nanmax(enrichment)], 'k--')
        axes[1].set_xlim(0, 1)
        axes[1].set_xlabel(f"Test {metric_label} cutoff")
        axes[1].set_ylabel("Enrichment over shuffle")
        fig.tight_layout()
        fig.savefig(path.join(output_folder, f"MINE_{your_model}_TestMetrics.pdf"))

    # plot linearity metrics and thresholds
    if np.any(model_scores >= test_score_thresh):
        fig = pl.figure()
        pl.scatter(mdata.model_lin_approx_scores, mdata.model_2nd_approx_scores, s=2)
        pl.plot([lax_thresh, lax_thresh], [-1, 1], 'k--')
        pl.plot([-1, 1], [sqr_thresh, sqr_thresh], 'k--')
        pl.xlim(-1, 1)
        pl.ylim(-1, 1)
        pl.xlabel("Linear approximation $R^2$")
        pl.ylabel("2nd order approximation $R^2$")
        fig.savefig(path.join(output_folder, f"MINE_{your_model}_LinearityMetrics.pdf"))

        # perform barcode clustering
        interpret_df = interpret_df[interpret_df["Fit"] == "Y"]
        barcode_labels = [ph for ph in predictor_columns] + ["Nonlinear"]
        barcode = np.hstack([(np.array(interpret_df[ph])=="Y")[:, None] for ph in predictor_columns])
        barcode = np.c_[barcode, (np.array(interpret_df["Linearity"])!="linear")[:, None]]
        df_barcode = pd.DataFrame(barcode, columns=barcode_labels)
        aggregate = ups.from_indicators(df_barcode)
        fig = pl.figure()
        up_set = ups.UpSet(aggregate, subset_size='count', min_subset_size=1, facecolor="C1", sort_by='cardinality',
                           sort_categories_by=None)
        axes_dict = up_set.plot(fig)
        axes_dict['intersections'].set_yscale('log')
        fig.savefig(path.join(output_folder, f"MINE_{your_model}_BarcodeUpsetPlot.pdf"))


if __name__ == '__main__':
    pass



"""
Module for easy running of MINE on user data
"""

import subprocess
import importlib.resources
import sys

def main_gui():
    with importlib.resources.path("neuro_mine.scripts", "app_gui.py") as script_path:
        subprocess.run(["python", str(script_path)])

def main():
    with importlib.resources.path("neuro_mine.scripts", "process_csv.py") as script_path:
        if len(sys.argv) > 1:
            subprocess.run(["python", str(script_path)] + sys.argv[1:])
        else:
            subprocess.run(["python", str(script_path)])